import os
from pathlib import Path
from types import LambdaType

import pytest

import autofit as af
from autoconf.conf import with_config
from autofit.mock.mock import MockSearch
from autofit.non_linear.analysis.multiprocessing import AnalysisPool
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.paths.sub_directory_paths import SubDirectoryPaths


class Analysis(af.Analysis):
    def __init__(self):
        self.did_visualise = False
        self.did_profile = False

    def log_likelihood_function(self, instance):
        return -1

    def visualize(
            self,
            paths: AbstractPaths,
            instance,
            during_analysis
    ):
        self.did_visualise = True
        os.makedirs(
            paths.image_path
        )
        open(f"{paths.image_path}/image.png", "w+").close()

    def profile_log_likelihood_function(
            self,
            paths: AbstractPaths,
            instance
    ):
        self.did_profile = True


def test_visualise():
    analysis_1 = Analysis()
    analysis_2 = Analysis()

    (analysis_1 + analysis_2).visualize(
        af.DirectoryPaths(), None, None
    )

    assert analysis_1.did_visualise is True
    assert analysis_2.did_visualise is True


def test__profile_log_likelihood():
    analysis_1 = Analysis()
    analysis_2 = Analysis()

    (analysis_1 + analysis_2).profile_log_likelihood_function(
        af.DirectoryPaths(), None,
    )

    assert analysis_1.did_profile is True
    assert analysis_2.did_profile is True


def test_make_result():
    analysis_1 = Analysis()
    analysis_2 = Analysis()

    result = (analysis_1 + analysis_2).make_result(
        samples=None, model=None, search=None
    )

    assert len(result) == 2


def test_add_analysis():
    assert (Analysis() + Analysis()).log_likelihood_function(
        None
    ) == -2


@pytest.mark.parametrize(
    "number, first, second",
    [
        (3, 2, 1),
        (4, 2, 2),
        (5, 3, 2),
        (6, 3, 3),
        (7, 4, 3),
    ]
)
def test_analysis_pool(
        number,
        first,
        second
):
    pool = AnalysisPool(
        number * [Analysis()],
        2
    )

    process_1, process_2 = pool.processes

    assert len(process_1.analyses) == first
    assert len(process_2.analyses) == second


@with_config(
    "general", "analysis", "n_cores",
    value=2
)
@pytest.mark.parametrize(
    "number",
    list(range(1, 3))
)
def test_two_cores(number):
    analysis = Analysis()
    for _ in range(number - 1):
        analysis += Analysis()
    assert analysis.log_likelihood_function(
        None
    ) == -number


def test_still_flat():
    analysis = (Analysis() + Analysis()) + Analysis()

    assert len(analysis) == 3

    analysis = Analysis() + (Analysis() + Analysis())

    assert len(analysis) == 3


def test_sum_analyses():
    combined = sum(
        Analysis()
        for _ in range(5)
    )
    assert len(combined) == 5


@pytest.fixture(
    name="search"
)
def make_search():
    return MockSearch(
        "search_name"
    )


def test_child_paths(
        search
):
    paths = search.paths
    sub_paths = SubDirectoryPaths(
        paths,
        analysis_name="analysis_0"
    )
    assert sub_paths.output_path == f"{paths.output_path}/analysis_0"


@pytest.fixture(
    name="multi_analysis"
)
def make_multi_analysis():
    return Analysis() + Analysis()


@pytest.fixture(
    name="multi_search"
)
def make_multi_search(
        search,
        multi_analysis
):
    search.paths.remove_files = False

    search.fit(
        af.Model(
            af.Gaussian
        ),
        multi_analysis
    )
    search.paths.save_all({}, {}, [])
    return search


@with_config(
    "general",
    "output",
    "remove_files",
    value=False
)
def test_visualise(
        multi_search,
        multi_analysis
):
    multi_analysis.visualize(
        multi_search.paths,
        af.Gaussian(),
        True
    )
    search_path = Path(multi_search.paths.output_path)
    assert search_path.exists()
    assert (search_path / "analyses/analysis_0/image/image.png").exists()
    assert (search_path / "analyses/analysis_1/image/image.png").exists()


def test_set_number_of_cores(
        multi_analysis
):
    multi_analysis.n_cores = 1
    assert isinstance(
        multi_analysis._log_likelihood_function,
        LambdaType
    )
    multi_analysis.n_cores = 2
    assert isinstance(
        multi_analysis._log_likelihood_function,
        AnalysisPool
    )
    multi_analysis.n_cores = 1
    assert isinstance(
        multi_analysis._log_likelihood_function,
        LambdaType
    )



import numpy as np
import pytest
from scipy import stats

import autofit.mapper.variable
from autofit import graphical as graph


@pytest.fixture(name="norm")
def make_norm():
    return stats.norm(loc=0, scale=1.0)


@pytest.fixture(autouse=True)
def set_seed():
    np.random.seed(1)


@pytest.fixture(name="prior_norm")
def make_prior_norm():
    return stats.norm(loc=0, scale=10.0)


@pytest.fixture(name="prior")
def make_prior(prior_norm):
    def prior(x):
        return prior_norm.logpdf(x).sum()

    return prior


def linear(x, a, b):
    return x.dot(a) + b


def linear_jacobian(x, a, b):
    z = np.matmul(x, a) + np.expand_dims(b, -2)

    (n, m) = z.shape
    d = np.shape(x)[1]
    i, j, k = np.indices((n, m, d))

    jac_x = np.zeros((n, m, n, d))
    jac_x[i, j, i, k] = a[k, j]
    jac_a = np.zeros((n, m, d, m))
    jac_a[i, j, k, j] = x[i, k]
    jac_b = np.repeat(np.eye(m)[None, ...], n, axis=0)
    jacs = (jac_x, jac_a, jac_b)

    return z, jacs


@pytest.fixture(name="obs")
def make_obs():
    return autofit.mapper.variable.Plate(name="obs")


@pytest.fixture(name="features")
def make_features():
    return autofit.mapper.variable.Plate(name="features")


@pytest.fixture(name="dims")
def make_dims():
    return autofit.mapper.variable.Plate(name="dims")


@pytest.fixture(name="x_")
def make_x_(obs, features):
    return autofit.mapper.variable.Variable("x", obs, features)


@pytest.fixture(name="a_")
def make_a_(features, dims):
    return autofit.mapper.variable.Variable("a", features, dims)


@pytest.fixture(name="b_")
def make_b_(dims):
    return autofit.mapper.variable.Variable("b", dims)


@pytest.fixture(name="z_")
def make_z_(obs, dims):
    return autofit.mapper.variable.Variable("z", obs, dims)


@pytest.fixture(name="y_")
def make_y_(obs, dims):
    return autofit.mapper.variable.Variable("y", obs, dims)


@pytest.fixture(name="linear_factor")
def make_linear_factor(x_, a_, b_, z_):
    return graph.Factor(linear, x_, a_, b_, factor_out=z_)


@pytest.fixture(name="linear_factor_jac")
def make_linear_factor_jac(x_, a_, b_, z_):
    return graph.Factor(
        linear, x_, a_, b_, factor_jacobian=linear_jacobian, factor_out=z_
    )


@pytest.fixture(name="prior_a")
def make_prior_a(prior, a_):
    return graph.Factor(prior, a_)


@pytest.fixture(name="prior_b")
def make_prior_b(prior, b_):
    return graph.Factor(prior, b_)


@pytest.fixture(name="likelihood_factor")
def make_likelihood_factor(likelihood, z_, y_):
    return graph.Factor(likelihood, z_, y_)



import pytest

from autofit.database.migration import SessionWrapper


@pytest.fixture(
    autouse=True
)
def create_table(
        session
):
    session.execute(
        "CREATE TABLE test (id INTEGER PRIMARY KEY)"
    )


def test_run_migration(
        migrator,
        session,
        revision_2
):
    migrator.migrate(
        session
    )
    assert len(list(
        session.execute(
            "SELECT * FROM test"
        )
    )) == 2

    assert SessionWrapper(
        session
    ).revision_id == revision_2


def test_apply_twice(
        migrator,
        session
):
    for _ in range(2):
        migrator.migrate(
            session
        )
    assert len(list(
        session.execute(
            "SELECT * FROM test"
        )
    )) == 2


def test_run_partial_migration(
        migrator,
        session,
        revision_1,
        revision_2
):
    wrapper = SessionWrapper(
        session
    )
    wrapper.revision_id = revision_1.id

    assert len(migrator.get_steps(revision_1.id)) == 1

    migrator.migrate(
        session
    )
    assert len(list(
        session.execute(
            "SELECT * FROM test"
        )
    )) == 1

    assert wrapper.revision_id == revision_2


def test_session_wrapper(session):
    wrapper = SessionWrapper(session)
    assert wrapper.revision_id is None

    wrapper.revision_id = "revision_id"
    assert wrapper.revision_id == "revision_id"


def test_table_exists(session):
    wrapper = SessionWrapper(session)
    assert wrapper.is_table is False

    _ = wrapper.revision_id
    assert wrapper.is_table is True



import numpy as np
import pytest
from scipy.optimize import approx_fprime

from autofit import graphical as graph
from autofit.mapper.variable import Variable, Plate
from autofit.messages.normal import NormalMessage


def log_sigmoid(x):
    return -np.log1p(np.exp(-x))


def log_phi(x):
    return -(x ** 2) / 2 - 0.5 * np.log(2 * np.pi)


def plus_two(x):
    return x + 2


@pytest.fixture(name="x")
def make_x():
    return Variable("x")


@pytest.fixture(name="y")
def make_y():
    return Variable("y")


@pytest.fixture(name="sigmoid")
def make_sigmoid(x):
    return graph.Factor(log_sigmoid, x)


@pytest.fixture(name="vectorised_sigmoid")
def make_vectorised_sigmoid(x):
    return graph.Factor(log_sigmoid, x)


@pytest.fixture(name="phi")
def make_phi(x):
    return graph.Factor(log_phi, x)


@pytest.fixture(name="compound")
def make_compound(sigmoid, phi):
    return sigmoid * phi


@pytest.fixture(name="plus")
def make_plus(x, y):
    return graph.Factor(plus_two, x, factor_out=y)


@pytest.fixture(name="flat_compound")
def make_flat_compound(plus, y, sigmoid):
    phi = graph.Factor(log_phi, y)
    return phi * plus * sigmoid


def test_factor_jacobian():
    shape = 4, 3
    z_ = Variable("z", *(Plate() for _ in shape))
    likelihood = NormalMessage(
        np.random.randn(*shape), np.random.exponential(size=shape)
    )
    likelihood_factor = likelihood.as_factor(z_)

    values = {z_: likelihood.sample()}
    fval, jval = likelihood_factor.func_jacobian(values)
    grad = jval.grad()
    ngrad = approx_fprime(
        values[z_].ravel(), lambda x: likelihood.logpdf(x.reshape(*shape)).sum(), 1e-8
    ).reshape(*shape)
    assert np.allclose(ngrad, grad[z_])


class TestFactorGraph:
    def test_names(self, sigmoid, phi, compound):
        assert sigmoid.name == "log_sigmoid"
        assert phi.name == "log_phi"
        # TODO: the below test was (log_sigmoid*log_phi).
        assert compound.name == "FactorGraph(Variable(x),)"

    def test_argument(self, x, sigmoid, phi, compound):
        values = {x: 5}
        assert sigmoid(values).log_value == -0.006715348489118068
        assert phi(values).log_value == -13.418938533204672
        assert compound(values).log_value == -13.42565388169379

    def test_deterministic_variable_value(self, flat_compound, x, y):
        value = flat_compound({x: 3})

        assert value.log_value == -13.467525884778414
        assert value.deterministic_values == {y: 5}

    @pytest.mark.parametrize("coefficient", [1, 2, 3, 4, 5])
    def test_jacobian(self, x, coefficient):
        factor = graph.Factor(lambda p: coefficient * p, x)

        assert factor.jacobian({x: 2}).grad()[x] == pytest.approx(coefficient)



import pytest

import autofit as af


@pytest.fixture(
    name="target_gaussian"
)
def make_target_gaussian():
    return af.PriorModel(
        af.Gaussian
    )


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.UniformPrior()


@pytest.fixture(
    name="source_gaussian"
)
def make_source_gaussian(prior):
    return af.PriorModel(
        af.Gaussian,
        centre=prior
    )


def test_simple(
        source_gaussian,
        target_gaussian,
        prior
):
    target_gaussian.take_attributes(
        source_gaussian
    )

    assert target_gaussian.centre == prior


def test_assertions(
        source_gaussian,
        target_gaussian
):
    target_gaussian.add_assertion(
        target_gaussian.centre <= target_gaussian.normalization
    )

    with pytest.raises(AssertionError):
        target_gaussian.take_attributes(
            source_gaussian
        )


def test_assertions_collection(
        source_gaussian,
        target_gaussian
):
    target_gaussian.add_assertion(
        target_gaussian.centre <= target_gaussian.normalization
    )

    target_collection = af.Collection(
        gaussian=target_gaussian
    )
    source_collection = af.Collection(
        gaussian=source_gaussian
    )

    with pytest.raises(AssertionError):
        target_collection.take_attributes(
            source_collection
        )


def test_in_collection(
        source_gaussian,
        target_gaussian,
        prior
):
    target = af.CollectionPriorModel(
        gaussian=target_gaussian
    )
    source = af.CollectionPriorModel(
        gaussian=source_gaussian
    )
    target.take_attributes(
        source
    )

    assert target.gaussian.centre == prior


def test_tuple(
        source_gaussian,
        target_gaussian,
        prior
):
    source_gaussian.centre = (prior, 1.0)
    target_gaussian.take_attributes(
        source_gaussian
    )

    assert target_gaussian.centre == (prior, 1.0)


def test_tuple_prior(
        source_gaussian,
        target_gaussian,
        prior
):
    source_gaussian.centre = (prior, 1.0)
    target_gaussian.centre = af.TuplePrior()
    target_gaussian.take_attributes(
        source_gaussian
    )

    assert target_gaussian.centre == (prior, 1.0)


def test_tuple_in_instance(
        target_gaussian,
        prior
):
    # noinspection PyTypeChecker
    source_gaussian = af.Gaussian(
        centre=(prior, 1.0)
    )
    target_gaussian.take_attributes(
        source_gaussian
    )

    assert target_gaussian.centre == (prior, 1.0)


def test_tuple_in_collection(
        source_gaussian,
        target_gaussian,
        prior
):
    source_gaussian.centre = (prior, 1.0)

    source = af.CollectionPriorModel(
        gaussian=source_gaussian
    )
    target = af.CollectionPriorModel(
        gaussian=target_gaussian
    )

    target.take_attributes(source)
    assert target.gaussian.centre == (prior, 1.0)


def test_tuple_in_instance_in_collection(
        target_gaussian,
        prior
):
    # noinspection PyTypeChecker
    source_gaussian = af.Gaussian(
        centre=(prior, 1.0)
    )

    source = af.CollectionPriorModel(
        gaussian=source_gaussian
    )
    target = af.CollectionPriorModel(
        gaussian=target_gaussian
    )

    target.take_attributes(source)
    assert target.gaussian.centre == (prior, 1.0)


def test_source_is_dict(
        source_gaussian,
        target_gaussian,
        prior
):
    source = dict(
        gaussian=source_gaussian
    )
    target = af.CollectionPriorModel(
        gaussian=target_gaussian
    )
    target.take_attributes(source)

    assert target.gaussian.centre == prior


def test_target_is_dict(
        source_gaussian,
        target_gaussian,
        prior
):
    source = af.CollectionPriorModel(
        collection=af.CollectionPriorModel(
            gaussian=source_gaussian
        )
    )
    target = af.CollectionPriorModel(
        collection=dict(
            gaussian=target_gaussian
        )
    )
    target.take_attributes(source)

    assert target.collection.gaussian.centre == prior


def test_missing_from_source(
        target_gaussian,
        prior
):
    target_gaussian.centre = prior

    target_gaussian.take_attributes(
        af.CollectionPriorModel()
    )
    assert target_gaussian.centre == prior


def test_unlabelled_in_collection(
        source_gaussian,
        target_gaussian,
        prior
):
    target = af.CollectionPriorModel(
        [target_gaussian]
    )
    source = af.CollectionPriorModel(
        [source_gaussian]
    )
    target.take_attributes(
        source
    )

    assert target[0].centre == prior


def test_passing_float(
        source_gaussian,
        target_gaussian
):
    source_gaussian.centre = 2.0
    target_gaussian.take_attributes(
        source_gaussian
    )

    assert target_gaussian.centre == 2.0


def test_missing_from_origin(
        target_gaussian
):
    target_gaussian.take_attributes(
        af.CollectionPriorModel()
    )


def test_limits(
        source_gaussian,
        target_gaussian
):
    source_gaussian.centre = af.GaussianPrior(
        mean=0,
        sigma=1,
        lower_limit=-1,
        upper_limit=1
    )
    target_gaussian.take_attributes(
        source_gaussian
    )
    assert target_gaussian.centre.lower_limit == -1
    assert target_gaussian.centre.upper_limit == 1


def test_tuples():
    centre = (0.0, 1.0)
    source = af.Model(
        af.Gaussian,
        centre=centre
    )
    target = af.Model(
        af.Gaussian
    )
    target.take_attributes(source)
    assert target.centre == centre



import numpy as np
import pytest

import autofit as af

from autofit.non_linear.grid import sensitivity as s

from autofit.mock.mock import MockSearch

x = np.array(range(10))


def image_function(instance: af.ModelInstance):
    image = instance.gaussian(x)
    if hasattr(instance, "perturbation"):
        image += instance.perturbation(x)
    return image


class Analysis(af.Analysis):

    def __init__(self, image: np.array):
        self.image = image

    def log_likelihood_function(self, instance):
        image = image_function(instance)
        return np.mean(np.multiply(-0.5, np.square(np.subtract(self.image, image))))


@pytest.fixture(
    name="perturbation_model"
)
def make_perturbation_model():
    return af.PriorModel(af.Gaussian)


@pytest.fixture(
    name="search"
)
def make_search():
    return MockSearch(return_sensitivity_results=True)


@pytest.fixture(
    name="sensitivity"
)
def make_sensitivity(
        perturbation_model,
        search
):
    # noinspection PyTypeChecker
    instance = af.ModelInstance()
    instance.gaussian = af.Gaussian()
    return s.Sensitivity(
        simulation_instance=instance,
        base_model=af.Collection(
            gaussian=af.PriorModel(af.Gaussian)
        ),
        perturbation_model=perturbation_model,
        simulate_function=image_function,
        analysis_class=Analysis,
        search=search,
        number_of_steps=2,
    )


@pytest.fixture(
    name="job"
)
def make_job(
        perturbation_model,
        search
):
    instance = af.ModelInstance()
    instance.gaussian = af.Gaussian()
    base_instance = instance
    instance.perturbation = af.Gaussian()
    image = image_function(instance)
    # noinspection PyTypeChecker
    return s.Job(
        model=af.Collection(
            gaussian=af.PriorModel(af.Gaussian)
        ),
        perturbation_model=af.PriorModel(af.Gaussian),
        base_instance=base_instance,
        perturbation_instance=instance,
        analysis=Analysis(image),
        search=search,
        number=1
    )



import numpy as np
import pytest
from scipy import integrate

from autofit.messages import transform
from autofit.messages.beta import BetaMessage
from autofit.messages.gamma import GammaMessage
from autofit.messages.normal import (
    NormalMessage,
    UniformNormalMessage,
    LogNormalMessage,
    MultiLogitNormalMessage,
)
from autofit.messages.transform import numerical_jacobian


def check_dist_norm(dist):
    norm, err = integrate.quad(dist.pdf, *dist._support[0])
    assert norm == pytest.approx(1, abs=err), dist


def check_dist_norms(dist):
    vals, err = integrate.quad_vec(
        lambda x: dist.pdf(np.full(dist.shape, x)), *dist._support[0]
    )
    assert np.allclose(vals, 1, atol=err), dist


def check_log_normalisation(ms):
    m1, *m2s = ms
    A = np.exp(m1.log_normalisation(*m2s))

    # Calculate normalisation numerically
    i12, ierr = integrate.quad(
        lambda x: np.exp(sum(m.logpdf(x) for m in ms)), *m1._support[0]
    )

    # verify within tolerance
    assert np.abs(A - i12) < ierr < 1e-1, m1


def check_numerical_gradient_hessians(message, x=None):
    x = message.sample() if x is None else x

    res = message.logpdf_gradient(x)
    nres = message.numerical_logpdf_gradient(x)
    for i, (x1, x2) in enumerate(zip(res, nres)):
        assert np.allclose(x1, x2, rtol=1e-3, atol=1e-2), (i, x1, x2, message)

    res = message.logpdf_gradient_hessian(x)
    nres = message.numerical_logpdf_gradient_hessian(x)
    for i, (x1, x2) in enumerate(zip(res, nres)):
        assert np.allclose(x1, x2, rtol=1e-3, atol=1e-2), (i, x1, x2, message)


def test_message_norm():
    messages = [
        tuple(map(NormalMessage, [0.5, 0.1], [0.2, 0.3])),
        tuple(map(NormalMessage, [0.5, 0.1, -0.5], [0.2, 0.3, 1.3])),
        tuple(map(GammaMessage, [0.5, 1.1], [0.2, 1.3])),
        tuple(map(GammaMessage, [0.5, 1.1, 2], [0.2, 1.3, 1])),
        tuple(map(BetaMessage, [2.0, 3.2, 1.5], [4.1, 2.3, 3])),
    ]
    for ms in messages:
        check_log_normalisation(ms)
        for m in ms:
            check_dist_norm(m)
            check_numerical_gradient_hessians(m)


N = NormalMessage
UN = UniformNormalMessage
SUN = UN.shifted(shift=0.3, scale=0.8)
LN = LogNormalMessage
MLN = MultiLogitNormalMessage
# test doubly transformed distributions
WN = NormalMessage.transformed(transform.log_transform).transformed(
    transform.exp_transform,
)


@pytest.mark.parametrize(
    "M, m, s, x",
    [
        (N, 1.0, 0.5, 0.3),
        (N, 1.0, 0.5, [0.3, 2.1]),
        (N, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], [0.1, 0.2, 0.3]),
        (N, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], [[0.1, 0.2, 0.3], [2.0, 1.0, -1]]),
        (UN, 1.0, 0.5, None),
        (UN, [0.1, 1.0, 2.0], [0.1, 0.5, 0.2], None),
        (SUN, 1.0, 0.5, None),
        (SUN, [0.1, 1.0, 2.0], [0.1, 0.5, 0.2], None),
        (LN, 1.0, 0.5, None),
        (LN, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], None),
        (LN, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], None),
        (MLN, [0.1, 1.0, 2.0], [0.1, 0.5, 0.2], None),
        (WN, 1.0, 0.5, 0.3),
        (WN, 1.0, 0.5, [0.3, 2.1]),
        (WN, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], [0.1, 0.2, 0.3]),
        (WN, [0.1, 1.0, 2.0], [2.0, 0.5, 3.0], [[0.1, 0.2, 0.3], [2.0, 1.0, -1]]),
    ],
)
def test_numerical_gradient_hessians(M, m, s, x):
    check_numerical_gradient_hessians(M(m, s), x)


def test_beta():
    a = b = np.r_[2.0, 3.2, 1.5]
    beta = BetaMessage(a, b[::-1])
    check_dist_norms(beta)

    betas = [BetaMessage(a, b) for a, b in (np.random.poisson(5, size=(10, 2)) + 1)]
    for b in betas:
        check_dist_norm(b)

    check_log_normalisation(betas)


def check_transforms(transform, x):
    y, logd, logd_grad, jac = transform.transform_det_jac(x)

    njac = numerical_jacobian(x, transform.transform)
    nlogd_grad = numerical_jacobian(x, transform.log_det)

    assert np.allclose(transform.inv_transform(y), x)
    assert np.allclose(jac.to_dense(), njac)
    assert np.isclose(logd.sum(), np.linalg.slogdet(njac)[1])
    assert np.allclose(nlogd_grad.sum(0), logd_grad)


def test_transforms():
    tests = [
        (transform.log_transform, np.r_[10, 2, 0.1]),
        (transform.exp_transform, np.r_[10, -2, 0.1]),
        (transform.logistic_transform, np.r_[0.22, 0.51, 0.1]),
        (transform.phi_transform, np.r_[0.22, 0.51, 0.1]),
        (transform.shifted_logistic(shift=11, scale=5.1), np.r_[11.1, 12, 16]),
    ]
    for args in tests:
        check_transforms(*args)


def test_multinomial_logit():
    mult_logit = transform.multinomial_logit_transform

    d = 3
    p = np.random.dirichlet(np.ones(d + 1))[:d]

    x, logd, logd_grad, jac = mult_logit.transform_det_jac(p)

    njac = numerical_jacobian(p, mult_logit.transform)
    nlogd_grad = numerical_jacobian(p, mult_logit.log_det)

    assert np.allclose(mult_logit.inv_transform(x), p)
    assert np.allclose(jac.to_dense(), njac)
    assert np.isclose(logd.sum(), np.linalg.slogdet(njac)[1])
    assert np.allclose(nlogd_grad.sum(0), logd_grad)

    n = 5

    ps = np.random.dirichlet(np.ones(d + 1), size=n)[:, :d]
    xs, logd, logd_grad, jac = mult_logit.transform_det_jac(ps)
    njac = numerical_jacobian(ps, mult_logit.transform).reshape(jac.shape)
    nlogd_grad = numerical_jacobian(ps, mult_logit.log_det)

    assert np.allclose(mult_logit.inv_transform(xs), ps)
    assert np.allclose(njac, jac.to_dense())
    assert np.isclose(
        logd.sum(), np.linalg.slogdet(njac.reshape(jac.lsize, jac.rsize))[1]
    )
    assert np.allclose(nlogd_grad.sum((0, 1)), logd_grad, 1e-5, 1e-3)

    assert np.allclose(xs[0], mult_logit.transform(ps[0]))
    assert np.allclose(logd[0], mult_logit.log_det(ps[0]))
    assert np.allclose(logd_grad[0], mult_logit.log_det_grad(ps[0])[1])


def test_normal_simplex():
    mult_logit = transform.MultinomialLogitTransform()
    NormalSimplex = NormalMessage.transformed(mult_logit)

    message = NormalSimplex([-1, 2], [0.3, 0.3])

    check_numerical_gradient_hessians(message, message.sample())

    def func(*p):
        return np.exp(message.factor(p)).prod()

    def simplex_lims(*args):
        return [0, 1 - sum(args)]

    # verify transformation normalises correctly
    res, err = integrate.nquad(func, [simplex_lims] * message.size)
    assert res == pytest.approx(1, rel=err)



import autofit as af
from autofit import graphical as g


def test_priors(
        hierarchical_factor
):
    assert len(hierarchical_factor.priors) == 2


def test_factors(
        hierarchical_factor
):
    assert len(hierarchical_factor.factors) == 1

    hierarchical_factor.add_drawn_variable(
        af.UniformPrior(
            lower_limit=0.0,
            upper_limit=1.0
        )
    )
    assert len(hierarchical_factor.factors) == 2


def test_factor(
        hierarchical_factor,
        prior
):
    factor = hierarchical_factor.factors[0]
    assert factor.distribution_model is hierarchical_factor
    assert factor.drawn_prior is prior


def test_graph(
        hierarchical_factor
):
    graph = g.FactorGraphModel(
        hierarchical_factor
    )
    assert len(graph.model_factors) == 1

    hierarchical_factor.add_drawn_variable(
        af.UniformPrior(
            lower_limit=0.0,
            upper_limit=1.0
        )
    )
    assert len(graph.model_factors) == 2



import pytest

import autofit as af
from autofit.mock.mock import MockAnalysis
from test_autofit.non_linear.grid.test_optimizer_grid_search import MockOptimizer


@pytest.fixture(
    name="search"
)
def make_search(session):
    return af.SearchGridSearch(
        search=MockOptimizer(
            session=session
        ),
        number_of_steps=2
    )


@pytest.fixture(
    autouse=True
)
def run_search(
        search,
        mapper
):
    search.fit(
        model=mapper,
        analysis=MockAnalysis(),
        grid_priors=[
            mapper.component.one_tuple.one_tuple_0,
            mapper.component.one_tuple.one_tuple_1,
        ]
    )


def test_save_result(
        search
):
    assert isinstance(
        search.paths.load_object(
            "result"
        ),
        af.GridSearchResult
    )


def test_aggregate(
        session
):
    aggregator = af.Aggregator(
        session
    )

    grid_searches = aggregator.grid_searches()
    assert len(
        grid_searches
    ) == 1
    assert len(
        grid_searches.children()
    ) > 0


def test_aggregate_completed(
        session
):
    session.commit()
    aggregator = af.Aggregator(
        session
    )
    aggregator = aggregator(
        aggregator.search.is_complete
    )
    grid_searches = aggregator.grid_searches()
    assert len(
        grid_searches
    ) == 1



import os
import shutil
from pathlib import Path

import pytest

import autofit as af
from autoconf.conf import output_path_for_test
from autofit.database import Fit
from autofit.database.sqlalchemy_ import sa

directory = Path(
    __file__
).parent

origin = Path(
    __file__
).parent / "database.sqlite"

copy = Path(
    __file__
).parent / "database_copy.sqlite"


@pytest.fixture(
    autouse=True
)
def manage_files():
    shutil.copy(
        origin,
        copy
    )
    yield
    os.remove(
        copy,

    )


def _check_migration():
    try:
        print(
            af.Aggregator.from_database(
                "database_copy.sqlite"
            ).fits
        )
    except sa.exc.OperationalError as e:
        raise AssertionError(
            "Migration steps are not complete"
        ) from e


@output_path_for_test(
    str(directory),
    remove=False
)
def test():
    """
    Raises an exception if changes to autofit/database/migration/steps.py
    are required to migrate old database files correctly.
    """
    _check_migration()


@output_path_for_test(
    str(directory),
    remove=False
)
def test_fail():
    Fit.random_column = sa.Column(
        sa.String
    )

    with pytest.raises(
            AssertionError
    ):
        _check_migration()



import zipfile
from os import listdir
from pathlib import Path

import pytest

import autofit as af
from autoconf.conf import output_path_for_test
from autofit import Gaussian
from autofit import conf
from autofit.database import Fit
from autofit.tools.update_identifiers import (
    update_directory_identifiers,
    update_database_identifiers,
    update_identifiers_from_dict
)

output_directory = Path(
    __file__
).parent / "output"


@output_path_for_test(
    output_directory,
)
def test_directory(
        old_directory_paths
):
    old_directory_paths.save_all()

    assert listdir(
        output_directory / "name"
    ) == [
               old_directory_paths.identifier
           ]

    conf.instance["general"]["output"]["identifier_version"] = 3
    update_directory_identifiers(
        output_directory
    )

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier != old_directory_paths.identifier
    assert suffix == "zip"


@pytest.fixture(
    name="old_directory_paths"
)
def make_old_directory_paths():
    conf.instance["general"]["output"]["identifier_version"] = 1
    search = af.DynestyStatic(
        name="name"
    )
    search.paths.model = af.PriorModel(
        af.Gaussian
    )
    return search.paths


@output_path_for_test(
    output_directory
)
def test_update_identifiers_from_dict_no_change():
    search = af.DynestyStatic(
        name="name"
    )
    search.paths.model = af.PriorModel(
        af.Gaussian
    )
    old_directory_paths = search.paths

    old_directory_paths.save_all()
    old_directory_paths.zip_remove()

    update_identifiers_from_dict(
        output_directory,
        {}
    )

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier == old_directory_paths.identifier
    assert suffix == "zip"


@output_path_for_test(
    output_directory
)
def test_update_identifiers_from_dict():
    search = af.DynestyStatic(
        name="name"
    )
    search.paths.model = af.PriorModel(
        af.Gaussian
    )
    old_directory_paths = search.paths

    initial_length = len(
        old_directory_paths._identifier.hash_list
    )

    old_directory_paths.save_all()
    old_directory_paths.zip_remove()

    update_identifiers_from_dict(
        output_directory,
        {
            "normalization": "magnitude"
        }
    )

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier != old_directory_paths.identifier
    assert suffix == "zip"

    unzipped = output_directory / "unzipped"
    with zipfile.ZipFile(output_directory / "name" / filename, "r") as f:
        f.extractall(unzipped)

    with open(
            unzipped / ".identifier"
    ) as f:
        lines = f.read().split("\n")
        assert "normalization" not in lines
        assert "magnitude" in lines

    assert len(lines) == initial_length


@output_path_for_test(
    output_directory,
)
def test_zipped_no_change(
        old_directory_paths
):
    old_directory_paths.save_all()
    old_directory_paths.zip_remove()

    update_directory_identifiers(
        output_directory
    )

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier == old_directory_paths.identifier
    assert suffix == "zip"


@output_path_for_test(
    output_directory,
)
def test_zipped(
        old_directory_paths
):
    old_directory_paths.save_all()
    old_directory_paths.zip_remove()

    assert listdir(
        output_directory / "name"
    ) == [
               f"{old_directory_paths.identifier}.zip"
           ]

    conf.instance["general"]["output"]["identifier_version"] = 3
    update_directory_identifiers(
        output_directory
    )

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier != old_directory_paths.identifier
    assert suffix == "zip"


def test_database(session):
    conf.instance["general"]["output"]["identifier_version"] = 1
    search = af.DynestyStatic(
        name="name",
        session=session
    )
    search.paths.model = af.PriorModel(
        af.Gaussian
    )
    search.paths.save_all(
        search_config_dict=search.config_dict_search,
        info={},
        pickle_files=[]
    )

    fit, = Fit.all(
        session=session
    )
    assert fit.id == search.paths.identifier

    conf.instance["general"]["output"]["identifier_version"] = 3
    update_database_identifiers(
        session
    )

    fit, = Fit.all(
        session=session
    )
    assert fit.id != search.paths.identifier



import pytest

import autofit as af
from autofit.exc import PriorLimitException


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.GaussianPrior(
        mean=3.0,
        sigma=5.0,
        lower_limit=0.0
    )


def test_intrinsic_lower_limit(prior):
    with pytest.raises(
            PriorLimitException
    ):
        prior.value_for(0.0)


def test_prior_factor(prior):
    prior.factor(1.0)

    with pytest.raises(
            PriorLimitException
    ):
        prior.factor(-1.0)


def test_optional(prior):
    prior.value_for(
        0.0,
        ignore_prior_limits=True
    )


@pytest.fixture(
    name="model"
)
def make_model(prior):
    return af.Model(
        af.Gaussian,
        centre=prior
    )


def test_vector_from_unit_vector(model):
    with pytest.raises(
            PriorLimitException
    ):
        model.vector_from_unit_vector([
            0, 0, 0
        ])


def test_vector_ignore_limits(model):
    model.vector_from_unit_vector(
        [0, 0, 0],
        ignore_prior_limits=True
    )


@pytest.mark.parametrize(
    "prior",
    [
        af.LogUniformPrior(),
        af.UniformPrior(),
        af.GaussianPrior(
            mean=0,
            sigma=1,
            lower_limit=0.0,
            upper_limit=1.0,
        )
    ]
)
@pytest.mark.parametrize(
    "value",
    [-1.0, 2.0]
)
def test_all_priors(
        prior,
        value
):
    with pytest.raises(
            PriorLimitException
    ):
        prior.value_for(value)

    prior.value_for(
        value,
        ignore_prior_limits=True
    )



import dill

import autofit as af
import autofit.graphical as ep
from test_autofit.graphical.gaussian.model import Gaussian, Analysis


def test_pickle():
    prior_model = af.Model(
        Gaussian
    )
    analysis_factor = ep.AnalysisFactor(
        prior_model,
        analysis=Analysis(
            x=1,
            y=2
        ),
    )

    analysis_factor = dill.loads(
        dill.dumps(analysis_factor)
    )

    assert isinstance(
        analysis_factor,
        ep.AnalysisFactor
    )



import pytest

import autofit as af
from autofit import database as db


@pytest.fixture(
    name="gaussian_0"
)
def make_gaussian_0(
        session
):
    gaussian_0 = db.Fit(
        id="gaussian_0",
        instance=af.Gaussian(
            centre=1
        ),
        info={"info": 1},
        is_complete=True,
        unique_tag="zero"
    )
    session.add(gaussian_0)
    session.commit()
    return gaussian_0


def test_order_by(
        aggregator,
        gaussian_1,
        gaussian_2,
        gaussian_0
):
    assert aggregator.order_by(
        aggregator.search.unique_tag
    ) == [gaussian_1, gaussian_2, gaussian_0]


def test_reversed(
        aggregator,
        gaussian_1,
        gaussian_2,
        gaussian_0
):
    assert aggregator.order_by(
        aggregator.search.unique_tag,
        reverse=True
    ) == [gaussian_0, gaussian_2, gaussian_1]


def test_boolean(
        aggregator,
        gaussian_1,
        gaussian_2
):
    assert aggregator.order_by(
        aggregator.search.is_complete
    ) == [gaussian_2, gaussian_1]
    assert aggregator.order_by(
        aggregator.search.is_complete,
        reverse=True
    ) == [gaussian_1, gaussian_2]


def test_combined(
        aggregator,
        gaussian_1,
        gaussian_2,
        gaussian_0
):
    assert aggregator.order_by(
        aggregator.search.is_complete
    ).order_by(
        aggregator.search.unique_tag
    ) == [gaussian_2, gaussian_1, gaussian_0]



from autofit.graphical import AnalysisFactor, PriorFactor


def test_non_trivial_results(
        non_trivial_model
):
    results_text = non_trivial_model.graph.make_results_text(
        non_trivial_model.mean_field_approximation()
    )
    assert results_text == """PriorFactors

PriorFactor0 (AnalysisFactor1.sigma)                                                      0.5
PriorFactor1 (AnalysisFactor1.normalization)                                              0.5
PriorFactor2 (AnalysisFactor0.centre, AnalysisFactor1.centre)                             0.5
PriorFactor3 (AnalysisFactor0.sigma)                                                      0.5
PriorFactor4 (AnalysisFactor0.normalization)                                              0.5

AnalysisFactors

AnalysisFactor0

centre (AnalysisFactor1.centre, PriorFactor2)                                             0.5
normalization (PriorFactor4)                                                              0.5
sigma (PriorFactor3)                                                                      0.5

AnalysisFactor1

centre (AnalysisFactor0.centre, PriorFactor2)                                             0.5
normalization (PriorFactor1)                                                              0.5
sigma (PriorFactor0)                                                                      0.5"""


def test_non_trivial_info(
        non_trivial_model
):
    info = non_trivial_model.graph.info
    assert info == """PriorFactors

PriorFactor0 (AnalysisFactor1.sigma)                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor1 (AnalysisFactor1.normalization)                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor2 (AnalysisFactor0.centre, AnalysisFactor1.centre)                             UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor3 (AnalysisFactor0.sigma)                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor4 (AnalysisFactor0.normalization)                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactors

AnalysisFactor0

centre (AnalysisFactor1.centre, PriorFactor2)                                             UniformPrior, lower_limit = 0.0, upper_limit = 1.0
normalization (PriorFactor4)                                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
sigma (PriorFactor3)                                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactor1

centre (AnalysisFactor0.centre, PriorFactor2)                                             UniformPrior, lower_limit = 0.0, upper_limit = 1.0
normalization (PriorFactor1)                                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
sigma (PriorFactor0)                                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0"""


def test_factors_grouped_by_type(
        factor_graph
):
    factors_by_type = factor_graph.factors_by_type()

    assert len(factors_by_type) == 2
    assert len(factors_by_type[AnalysisFactor]) == 2
    assert len(factors_by_type[PriorFactor]) == 2


def test_make_results_text(
        factor_graph,
        factor_graph_model
):
    results_text = factor_graph.make_results_text(
        factor_graph_model.mean_field_approximation()
    )
    assert results_text == """PriorFactors

PriorFactor0 (AnalysisFactor1.one)                                                        0.5
PriorFactor1 (AnalysisFactor0.one)                                                        0.5

AnalysisFactors

AnalysisFactor0

one (PriorFactor1)                                                                        0.5

AnalysisFactor1

one (PriorFactor0)                                                                        0.5"""


def test_info_for_prior_factor(
        declarative_graph_output,
        prior_factor
):
    assert declarative_graph_output.info_for_prior_factor(
        prior_factor
    ) == "PriorFactor0 (AnalysisFactor1.one)                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0"


def test_info_for_analysis_factor(
        declarative_graph_output,
        analysis_factor
):
    info = declarative_graph_output.info_for_analysis_factor(
        analysis_factor
    )
    assert info == """AnalysisFactor0

one (PriorFactor1)                                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0"""


def test_related_factors(
        factor_graph,
        prior_factor
):
    assert len(factor_graph.related_factors(
        list(prior_factor.variables)[0]
    )) == 2


def test_graph_info(
        factor_graph
):
    info = factor_graph.info
    assert info == """PriorFactors

PriorFactor0 (AnalysisFactor1.one)                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor1 (AnalysisFactor0.one)                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactors

AnalysisFactor0

one (PriorFactor1)                                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactor1

one (PriorFactor0)                                                                        UniformPrior, lower_limit = 0.0, upper_limit = 1.0"""



import logging
from abc import ABC

from autoconf import conf
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import CollectionPriorModel
from autofit.non_linear.analysis.multiprocessing import AnalysisPool
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result
from autofit.non_linear.samples import Samples

logger = logging.getLogger(
    __name__
)


class Analysis(ABC):
    """
    Protocol for an analysis. Defines methods that can or
    must be implemented to define a class that compute the
    likelihood that some instance fits some data.
    """

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        pass

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        pass

    def save_results_for_aggregator(self, paths: AbstractPaths, model: CollectionPriorModel,
                                    samples: Samples):
        pass

    def modify_before_fit(self, paths: AbstractPaths, model: AbstractPriorModel):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def modify_after_fit(self, paths: AbstractPaths, model: AbstractPriorModel, result: Result):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def make_result(self, samples, model, search):
        return Result(samples=samples, model=model, search=search)

    def profile_log_likelihood_function(self, paths: AbstractPaths, instance):
        """
        Overwrite this function for profiling of the log likelihood function to be performed every update of a 
        non-linear search.
        
        This behaves analogously to overwriting the `visualize` function of the `Analysis` class, whereby the user 
        fills in the project-specific behaviour of the profiling.
        
        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        """
        pass

    def __add__(
            self,
            other: "Analysis"
    ) -> "CombinedAnalysis":
        """
        Analyses can be added together. The resultant
        log likelihood function returns the sum of the
        underlying log likelihood functions.

        Parameters
        ----------
        other
            Another analysis class

        Returns
        -------
        A class that computes log likelihood based on both analyses
        """
        if isinstance(
                other,
                CombinedAnalysis
        ):
            return other + self
        return CombinedAnalysis(
            self, other
        )

    def __radd__(self, other):
        """
        Allows analysis to be used in sum
        """
        if other == 0:
            return self
        return self + other


class CombinedAnalysis(Analysis):
    def __init__(self, *analyses: Analysis):
        """
        Computes the summed log likelihood of multiple analyses
        applied to a single model.

        Either analyses are performed sequentially and summed,
        or they are mapped out to processes.

        If the number of cores is greater than one then the
        analyses are distributed across a number of processes
        equal to the number of cores.

        Parameters
        ----------
        analyses
        """
        self.analyses = analyses
        self._n_cores = None
        self._log_likelihood_function = None
        self.n_cores = conf.instance[
            "general"
        ][
            "analysis"
        ][
            "n_cores"
        ]

    @property
    def n_cores(self):
        return self._n_cores

    @n_cores.setter
    def n_cores(self, n_cores):
        self._n_cores = n_cores
        if self.n_cores > 1:
            analysis_pool = AnalysisPool(
                self.analyses,
                self.n_cores
            )
            self._log_likelihood_function = analysis_pool
        else:
            self._log_likelihood_function = lambda instance: sum(
                analysis.log_likelihood_function(
                    instance
                )
                for analysis in self.analyses
            )

    def log_likelihood_function(self, instance):
        return self._log_likelihood_function(instance)

    def _for_each_analysis(self, func, paths):
        """
        Convenience function to call an underlying function for each
        analysis with a paths object with an integer attached to the
        end.

        Parameters
        ----------
        func
            Some function of the analysis class
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        """
        for i, analysis in enumerate(self.analyses):
            child_paths = paths.for_sub_analysis(
                analysis_name=f"analyses/analysis_{i}"
            )
            func(child_paths, analysis)

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        def func(child_paths, analysis):
            analysis.save_attributes_for_aggregator(
                child_paths,
            )

        self._for_each_analysis(
            func,
            paths
        )

    def save_results_for_aggregator(
            self,
            paths: AbstractPaths,
            model: CollectionPriorModel,
            samples: Samples
    ):
        def func(child_paths, analysis):
            analysis.save_results_for_aggregator(
                child_paths,
                model,
                samples
            )

        self._for_each_analysis(
            func,
            paths
        )

    def visualize(
            self,
            paths: AbstractPaths,
            instance,
            during_analysis
    ):
        """
        Visualise the instance according to each analysis.

        Visualisation output is distinguished by using an integer suffix
        for each analysis path.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        during_analysis
            Is this visualisation during analysis?
        """

        def func(child_paths, analysis):
            analysis.visualize(
                child_paths,
                instance,
                during_analysis
            )

        self._for_each_analysis(
            func,
            paths
        )

    def profile_log_likelihood_function(
            self,
            paths: AbstractPaths,
            instance,
    ):
        """
        Profile the log likelihood function of the maximum likelihood model instance using each analysis.

        Profiling output is distinguished by using an integer suffix for each analysis path.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        """

        def func(child_paths, analysis):
            analysis.profile_log_likelihood_function(
                child_paths,
                instance,
            )

        self._for_each_analysis(
            func,
            paths
        )

    def make_result(
            self, samples, model, search
    ):
        return [analysis.make_result(samples, model, search) for analysis in self.analyses]

    def __len__(self):
        return len(self.analyses)

    def __add__(self, other: Analysis):
        """
        Adding anything to a CombinedAnalysis results in another
        analysis containing all underlying analyses (no combined
        analysis children)

        Parameters
        ----------
        other
            Some analysis

        Returns
        -------
        An overarching analysis
        """
        if isinstance(
                other,
                CombinedAnalysis
        ):
            return CombinedAnalysis(
                *self.analyses,
                *other.analyses
            )
        return CombinedAnalysis(
            *self.analyses,
            other
        )



import logging
import os
from pathlib import Path

import pytest

from autoconf.conf import output_path_for_test

import autofit as af
from autofit import database as m


@pytest.fixture(
    name="fit"
)
def query_fit(session, paths):
    fit, = m.Fit.all(session)
    return fit


output_path = str(
    Path(
        __file__
    ).parent / "temp"
)

@output_path_for_test(
    output_path
)
def test_create():
    m.open_database("test.sqlite")
    assert os.path.exists(
        output_path
    )


@output_path_for_test(
    output_path
)
def test_make_dirs():
    m.open_database(
        f"a/long/path.sqlite"
    )


@output_path_for_test(
    output_path
)
def test_create_postgres():
    try:
        m.open_database(
            "postgresql://autofit:autofit@localhost/autofit"
        )
    except Exception as e:
        logging.exception(e)
    assert not os.path.exists(
        output_path
    )


def test_incomplete(paths):
    assert paths.is_complete is False


def test_identifier(
        paths,
        fit
):
    assert fit.id == paths.identifier


def test_completion(
        paths,
        fit
):
    paths.completed()

    assert fit.is_complete
    assert paths.is_complete


def test_object(paths):
    gaussian = af.Gaussian(
        normalization=2.1
    )

    assert paths.is_object(
        "gaussian"
    ) is False

    paths.save_object(
        "gaussian",
        gaussian
    )

    assert paths.is_object(
        "gaussian"
    ) is True
    assert paths.load_object(
        "gaussian"
    ) == gaussian

    paths.remove_object(
        "gaussian"
    )
    assert paths.is_object(
        "gaussian"
    ) is False


def test_save_all(
        paths,
        fit
):
    paths.save_all({
        "key": "value"
    })

    assert fit.model is not None
    assert "search" in fit
    assert fit.info["key"] == "value"



import numpy as np
import pytest
from scipy import stats, linalg

import autofit.graphical.factor_graphs.transform as transform


def test_diagonal_from_dense():
    matrix = transform.DiagonalMatrix.from_dense(
        np.array([[4, 2], [3, 4]])
    ).to_diagonal()

    assert len(matrix.scale) == 2
    assert (matrix.scale == np.array([4, 4])).all()


@pytest.mark.parametrize(
    "bounds, result",
    [
        ([(1, 2), (1, 2)], [(3, 6), (2, 4)]),
        ([(np.inf, np.inf), (np.inf, np.inf)], [(np.inf, np.inf), (np.inf, np.inf)]),
        ([(-2, -1), (0, 2)], [(-6, -3), (0, 4)]),
    ],
)
def test_transform_bounds(bounds, result):
    matrix = transform.DiagonalMatrix(np.array([3, 2]))
    assert matrix.transform_bounds(bounds) == result


def test_cholesky_transform():
    d = 10
    A = stats.wishart(d, np.eye(d)).rvs()

    cho_factor = transform.CholeskyOperator(linalg.cho_factor(A))

    U = np.triu(cho_factor.U)
    iU = np.linalg.inv(U)

    b = np.random.rand(d)
    assert np.allclose(cho_factor * b, U @ b)
    assert np.allclose(b * cho_factor, b @ U)
    assert np.allclose(cho_factor.ldiv(b), iU @ b)
    assert np.allclose(b / cho_factor, b @ iU)

    b = np.random.rand(d, d)
    assert np.allclose(cho_factor * b, U @ b)
    assert np.allclose(b * cho_factor, b @ U)
    assert np.allclose(cho_factor.ldiv(b), iU @ b)
    assert np.allclose(b / cho_factor, b @ iU)

    b = np.random.rand(d, d + 1)
    assert np.allclose(cho_factor * b, U @ b)
    assert np.allclose(cho_factor.ldiv(b), iU @ b)

    b = np.random.rand(d + 1, d)
    assert np.allclose(b * cho_factor, b @ U)
    assert np.allclose(b / cho_factor, b @ iU)


def test_diagonal_transform():
    d = 3

    scale = np.random.rand(d)
    D = np.diag(scale)
    iD = np.diag(scale ** -1)
    diag_scale = transform.DiagonalMatrix(scale)

    b = np.random.rand(d)
    assert np.allclose(diag_scale * b, D @ b)
    assert np.allclose(b * diag_scale, b @ D)
    assert np.allclose(diag_scale.ldiv(b), iD @ b)
    assert np.allclose(b / diag_scale, b @ iD)

    b = np.random.rand(d, d)
    assert np.allclose(diag_scale * b, D @ b)
    assert np.allclose(b * diag_scale, b @ D)
    assert np.allclose(diag_scale.ldiv(b), iD @ b)
    assert np.allclose(b / diag_scale, b @ iD)

    b = np.random.rand(d, d + 1)
    assert np.allclose(diag_scale * b, D @ b)
    assert np.allclose(diag_scale.ldiv(b), iD @ b)

    b = np.random.rand(d + 1, d)
    assert np.allclose(b * diag_scale, b @ D)
    assert np.allclose(b / diag_scale, b @ iD)



import autofit as af
import numpy as np
import pytest
from autofit.mock.mock_model import (
    MockClassx2,
    MockClassx2Tuple,
    MockClassx2NoSuperScript,
    MockClassx3,
    MockClassx3TupleFloat,
    MockClassRelativeWidth,
    MockWithFloat,
    MockWithTuple
)
from autofit.mock.mock_model import MockChildTuplex2, MockChildTuplex3, MockChildTuple


@pytest.fixture(name="initial_model")
def make_initial_model():
    return af.PriorModel(MockClassx2)


class TestParamNames:
    def test_has_prior(self):
        prior_model = af.PriorModel(MockClassx2)
        assert "one" == prior_model.name_for_prior(prior_model.one)


class MockClassx2instance(MockClassx2):
    pass


class ExtendedMockClass(MockClassx2):
    def __init__(self, one, two, three):
        super().__init__(one, two)
        self.three = three


# noinspection PyUnresolvedReferences
class TestRegression:
    def test_set_tuple_instance(self):
        mm = af.ModelMapper()
        mm.mock_cls = MockChildTuplex2

        assert mm.prior_count == 4

        mm.mock_cls.tup_0 = 0.0
        mm.mock_cls.tup_1 = 0.0

        assert mm.prior_count == 2

    def test_get_tuple_instances(self):
        mm = af.ModelMapper()
        mm.mock_cls = MockChildTuplex2

        assert isinstance(mm.mock_cls.tup_0, af.Prior)
        assert isinstance(mm.mock_cls.tup_1, af.Prior)

    def test_tuple_parameter(self, mapper):
        mapper.with_float = MockWithFloat
        mapper.with_tuple = MockWithTuple

        assert mapper.prior_count == 3

        mapper.with_tuple.tup_0 = mapper.with_float.value

        assert mapper.prior_count == 2

    def test_parameter_name_ordering(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassRelativeWidth

        mm.one.one.id = mm.two.three.id + 1

        assert mm.model_component_and_parameter_names == [
            "one_two",
            "one_three",
            "two_one",
            "two_two",
            "two_three",
            "one_one",
        ]

    def test_parameter_name_list(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassRelativeWidth

        assert mm.parameter_names == ["one", "two", "three", "one", "two", "three"]

    def test_parameter_name_distinction(self):
        mm = af.ModelMapper()
        mm.ls = af.CollectionPriorModel(
            [
                af.PriorModel(MockClassRelativeWidth),
                af.PriorModel(MockClassRelativeWidth),
            ]
        )
        assert mm.model_component_and_parameter_names == [
            "ls_0_one",
            "ls_0_two",
            "ls_0_three",
            "ls_1_one",
            "ls_1_two",
            "ls_1_three",
        ]

    def test__parameter_labels(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassx2

        assert mm.parameter_labels == [
            "one_label",
            "two_label",
            "three_label",
            "one_label",
            "two_label",
        ]

    def test__superscripts(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassx2NoSuperScript

        assert mm.superscripts == ['r', 'r', 'r', 'two', 'two']

        model = af.Collection(group=mm)

        assert model.superscripts == ['r', 'r', 'r', "two", "two"]

    def test__superscript_overwrite_via_config(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassx2NoSuperScript
        mm.three = MockClassx3

        assert mm.superscripts_overwrite_via_config == ['r', 'r', 'r', "", "", "", "", ""]

    def test__parameter_labels_with_superscripts_latex(self):
        mm = af.ModelMapper()
        mm.one = MockClassRelativeWidth
        mm.two = MockClassx2NoSuperScript

        assert mm.parameter_labels_with_superscripts_latex == [
            r"$one_label^{\rm r}$",
            r"$two_label^{\rm r}$",
            r"$three_label^{\rm r}$",
            r"$one_label^{\rm two}$",
            r"$two_label^{\rm two}$",
        ]

    def test_name_for_prior(self):
        ls = af.CollectionPriorModel(
            [
                MockClassRelativeWidth(1, 2, 3),
                af.PriorModel(MockClassRelativeWidth),
            ]
        )
        assert ls.name_for_prior(ls[1].one) == "1_one"

    def test_tuple_parameter_float(self, mapper):
        mapper.with_float = MockWithFloat
        mapper.with_tuple = MockWithTuple

        mapper.with_float.value = 1.0

        assert mapper.prior_count == 2

        mapper.with_tuple.tup_0 = mapper.with_float.value

        assert mapper.prior_count == 1

        instance = mapper.instance_from_unit_vector([0.0])

        assert instance.with_float.value == 1
        assert instance.with_tuple.tup == (1.0, 0.0)


class TestModelingMapper:
    def test__argument_extraction(self):
        mapper = af.ModelMapper()
        mapper.mock_class = MockClassx2
        assert 1 == len(mapper.prior_model_tuples)

        assert len(mapper.prior_tuples_ordered_by_id) == 2

    def test_attribution(self):
        mapper = af.ModelMapper()

        mapper.mock_class = MockClassx2

        assert hasattr(mapper, "mock_class")
        assert hasattr(mapper.mock_class, "one")

    def test_tuple_arg(self):
        mapper = af.ModelMapper()

        mapper.mock_profile = MockClassx3TupleFloat

        assert 3 == len(mapper.prior_tuples_ordered_by_id)


class TestInstances:

    def test_attribute(self):
        mm = af.ModelMapper()
        mm.cls_1 = MockClassx2

        assert 1 == len(mm.prior_model_tuples)
        assert isinstance(mm.cls_1, af.PriorModel)

    def test__instance_from_unit_vector(self):
        mapper = af.ModelMapper(mock_cls=MockClassx2Tuple)

        model_map = mapper.instance_from_unit_vector([1.0, 1.0])

        assert model_map.mock_cls.one_tuple == (1.0, 2.0)

    def test__instance_from_vector(self):
        mapper = af.ModelMapper(mock_cls=MockClassx2Tuple)

        model_map = mapper.instance_from_vector([1.0, 0.5])

        assert model_map.mock_cls.one_tuple == (1.0, 0.5)

    def test_inheritance(self):
        mapper = af.ModelMapper(mock_cls=MockChildTuplex2)

        model_map = mapper.instance_from_unit_vector([1.0, 1.0, 1.0, 1.0])

        assert model_map.mock_cls.tup == (1.0, 1.0)

    def test__multiple_classes(self):

        mapper = af.ModelMapper(
            mock_child_cls_0=MockChildTuplex3,
            mock_child_cls_1=MockChildTuplex2,
            mock_child_cls_2=MockChildTuplex2,
            mock_child_tuple=MockChildTuple,
            mock_child_cls_3=MockChildTuplex3,
        )

        model_map = mapper.instance_from_unit_vector(
            [0.5 for _ in range(len(mapper.prior_tuples_ordered_by_id))]
        )

        assert isinstance(model_map.mock_child_cls_1, MockChildTuplex2)
        assert isinstance(model_map.mock_child_cls_2, MockChildTuplex2)
        assert isinstance(model_map.mock_child_tuple, MockChildTuple)

        assert isinstance(model_map.mock_child_cls_0, MockChildTuplex3)
        assert isinstance(
            model_map.mock_child_cls_3, MockChildTuplex3
        )

    def test__in_order_of_class_constructor(self):
        mapper = af.ModelMapper(mock_cls_0=MockChildTuplex2)

        model_map = mapper.instance_from_unit_vector([0.25, 0.5, 0.75, 1.0])

        assert model_map.mock_cls_0.tup == (0.25, 0.5)
        assert model_map.mock_cls_0.one == 1.5
        assert model_map.mock_cls_0.two == 2.0

        mapper = af.ModelMapper(
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuple,
            mock_cls_2=MockChildTuplex2,
        )

        model_map = mapper.instance_from_unit_vector(
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        )

        assert model_map.mock_cls_0.tup == (0.1, 0.2)
        assert model_map.mock_cls_0.one == 0.6
        assert model_map.mock_cls_0.two == 0.8

        assert model_map.mock_cls_1.tup == (0.5, 0.6)

        assert model_map.mock_cls_2.tup == (0.7, 0.8)
        assert model_map.mock_cls_2.one == 1.8
        assert model_map.mock_cls_2.two == 2.0

    def test__check_order_for_different_unit_values(self):

        mapper = af.ModelMapper(
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuple,
            mock_cls_2=MockChildTuplex2,
        )

        mapper.mock_cls_0.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.tup.tup_1 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.one = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.two = af.UniformPrior(0.0, 1.0)

        mapper.mock_cls_1.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_1.tup.tup_1 = af.UniformPrior(0.0, 1.0)

        mapper.mock_cls_2.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.tup.tup_1 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.one = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.two = af.UniformPrior(0.0, 1.0)

        model_map = mapper.instance_from_unit_vector(
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        )

        assert model_map.mock_cls_0.tup == (0.1, 0.2)
        assert model_map.mock_cls_0.one == 0.3
        assert model_map.mock_cls_0.two == 0.4

        assert model_map.mock_cls_1.tup == (0.5, 0.6)

        assert model_map.mock_cls_2.tup == (0.7, 0.8)
        assert model_map.mock_cls_2.one == 0.9
        assert model_map.mock_cls_2.two == 1.0

    def test__check_order_for_different_unit_values_and_set_priors_equal_to_one_another(
            self
    ):
        mapper = af.ModelMapper(
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuple,
            mock_cls_2=MockChildTuplex2,
        )

        mapper.mock_cls_0.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.tup.tup_1 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.one = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_0.two = af.UniformPrior(0.0, 1.0)

        mapper.mock_cls_1.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_1.tup.tup_1 = af.UniformPrior(0.0, 1.0)

        mapper.mock_cls_2.tup.tup_0 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.tup.tup_1 = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.one = af.UniformPrior(0.0, 1.0)
        mapper.mock_cls_2.two = af.UniformPrior(0.0, 1.0)

        mapper.mock_cls_0.one = mapper.mock_cls_0.two
        mapper.mock_cls_2.tup.tup_1 = mapper.mock_cls_1.tup.tup_1

        model_map = mapper.instance_from_unit_vector(
            [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        )

        assert model_map.mock_cls_0.tup == (0.2, 0.3)
        assert model_map.mock_cls_0.one == 0.4
        assert model_map.mock_cls_0.two == 0.4

        assert model_map.mock_cls_1.tup == (0.5, 0.6)

        assert model_map.mock_cls_2.tup == (0.7, 0.6)
        assert model_map.mock_cls_2.one == 0.8
        assert model_map.mock_cls_2.two == 0.9

    def test__instance_from_vector__check_order(self):
        mapper = af.ModelMapper(
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuple,
            mock_cls_2=MockChildTuplex2,
        )

        model_map = mapper.instance_from_vector(
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        )

        assert model_map.mock_cls_0.tup == (0.1, 0.2)
        assert model_map.mock_cls_0.one == 0.3
        assert model_map.mock_cls_0.two == 0.4

        assert model_map.mock_cls_1.tup == (0.5, 0.6)

        assert model_map.mock_cls_2.tup == (0.7, 0.8)
        assert model_map.mock_cls_2.one == 0.9
        assert model_map.mock_cls_2.two == 1.0

    def test__instance_from_prior_medians(self):
        mapper = af.ModelMapper(mock_cls_0=MockChildTuplex2)

        model_map = mapper.instance_from_prior_medians()

        model_2 = mapper.instance_from_unit_vector([0.5, 0.5, 0.5, 0.5])

        assert model_map.mock_cls_0.tup == model_2.mock_cls_0.tup == (0.5, 0.5)
        assert model_map.mock_cls_0.one == model_2.mock_cls_0.one == 1.0
        assert model_map.mock_cls_0.two == model_2.mock_cls_0.two == 1.0

        mapper = af.ModelMapper(
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuple,
            mock_cls_2=MockChildTuplex2,
        )

        model_map = mapper.instance_from_prior_medians()

        model_2 = mapper.instance_from_unit_vector(
            [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
        )

        assert model_map.mock_cls_0.tup == model_2.mock_cls_0.tup == (0.5, 0.5)
        assert model_map.mock_cls_0.one == model_2.mock_cls_0.one == 1.0
        assert model_map.mock_cls_0.two == model_2.mock_cls_0.two == 1.0

        assert model_map.mock_cls_1.tup == model_2.mock_cls_1.tup == (0.5, 0.5)

        assert model_map.mock_cls_2.tup == model_2.mock_cls_2.tup == (0.5, 0.5)
        assert model_map.mock_cls_2.one == model_2.mock_cls_2.one == 1.0
        assert model_map.mock_cls_2.two == model_2.mock_cls_2.two == 1.0

    def test__from_prior_medians__one_model__set_one_parameter_to_another(self):
        mapper = af.ModelMapper(mock_cls_0=MockChildTuplex2)

        mapper.mock_cls_0.one = mapper.mock_cls_0.two

        model_map = mapper.instance_from_prior_medians()

        model_2 = mapper.instance_from_unit_vector([0.5, 0.5, 0.5])

        assert model_map.mock_cls_0.tup == model_2.mock_cls_0.tup == (0.5, 0.5)
        assert model_map.mock_cls_0.one == model_2.mock_cls_0.one == 1.0
        assert model_map.mock_cls_0.two == model_2.mock_cls_0.two == 1.0

    def test_log_prior_list_from_vector(self):
        mapper = af.ModelMapper()
        mapper.mock_class = af.PriorModel(MockClassx2)
        mapper.mock_class.one = af.GaussianPrior(mean=1.0, sigma=2.0)
        mapper.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        log_prior_list = mapper.log_prior_list_from_vector(vector=[0.0, 5.0])

        assert log_prior_list == [0.125, 0.2]

    def test_random_unit_vector_within_limits(self):
        mapper = af.ModelMapper()
        mapper.mock_class = af.PriorModel(MockClassx2)

        np.random.seed(1)

        assert mapper.random_unit_vector_within_limits(
            lower_limit=0.0, upper_limit=1.0
        ) == pytest.approx([0.41702, 0.720324], 1.0e-4)

        assert mapper.random_unit_vector_within_limits(
            lower_limit=0.2, upper_limit=0.8
        ) == pytest.approx([0.200068, 0.38140], 1.0e-4)

    def test_random_vector_from_prior_within_limits(self):
        np.random.seed(1)

        mapper = af.ModelMapper()
        mapper.mock_class = af.PriorModel(MockClassx2)

        vector = mapper.random_vector_from_priors_within_limits(
            lower_limit=0.499999, upper_limit=0.500001
        )

        assert vector == pytest.approx([0.5, 1.0], 1.0e-4)

        vector = mapper.random_vector_from_priors_within_limits(
            lower_limit=0.899999, upper_limit=0.900001
        )

        assert vector == pytest.approx([0.9, 1.8], 1.0e-4)

        vector = mapper.random_vector_from_priors_within_limits(
            lower_limit=0.2, upper_limit=0.3
        )

        assert vector == pytest.approx([0.21467, 0.4184677], 1.0e-4)

    def test_random_vector_from_prior(self):
        mapper = af.Collection(
            mock_class=af.Model(MockClassx2)
        )

        np.random.seed(1)

        assert mapper.random_vector_from_priors == pytest.approx(
            [0.41702, 1.44064], 1.0e-4
        )
        assert mapper.random_vector_from_priors == pytest.approx(
            [0.00011437, 0.6046651], 1.0e-4
        )

        # By default, this seeded random will draw a value < -0.15, which is below the lower limit below. This
        # test ensures that this value is resampled to the next draw, which is above 0.15

        mapper = af.Collection(
            mock_class=af.Model(
                MockClassx2,
                one=af.UniformPrior(
                    lower_limit=0.15,
                    upper_limit=1.0
                )
            )
        )

        mapper.mock_class.one.lower_limit = 0.15

        assert mapper.random_vector_from_priors == pytest.approx(
            [0.27474, 0.1846771], 1.0e-4
        )

    def test_vector_from_prior_medians(self):
        mapper = af.ModelMapper()
        mapper.mock_class = af.PriorModel(MockClassx2)

        assert mapper.physical_values_from_prior_medians == [0.5, 1.0]


class TestUtility:
    def test_prior_prior_model_dict(self):
        mapper = af.ModelMapper(mock_class=MockClassx2)

        assert len(mapper.prior_prior_model_dict) == 2
        assert (
                mapper.prior_prior_model_dict[mapper.prior_tuples_ordered_by_id[0][1]].cls
                == MockClassx2
        )
        assert (
                mapper.prior_prior_model_dict[mapper.prior_tuples_ordered_by_id[1][1]].cls
                == MockClassx2
        )

    def test_name_for_prior(self):
        mapper = af.ModelMapper(mock_class=MockClassx2)

        assert mapper.name_for_prior(mapper.priors[0]) == "mock_class_one"
        assert mapper.name_for_prior(mapper.priors[1]) == "mock_class_two"


class TestPriorReplacement:
    def test_prior_replacement(self):
        mapper = af.ModelMapper(mock_class=MockClassx2)
        result = mapper.mapper_from_gaussian_tuples([(10, 3), (5, 3)])

        assert isinstance(result.mock_class.one, af.GaussianPrior)
        assert {
                   prior.id for prior in mapper.priors
               } == {
                   prior.id for prior in result.priors
               }

    def test_replace_priors_with_gaussians_from_tuples(self):
        mapper = af.ModelMapper(mock_class=MockClassx2)
        result = mapper.mapper_from_gaussian_tuples([(10, 3), (5, 3)])

        assert isinstance(result.mock_class.one, af.GaussianPrior)

    def test_replacing_priors_for_profile(self):
        mapper = af.ModelMapper(mock_class=MockClassx3TupleFloat)
        result = mapper.mapper_from_gaussian_tuples([(10, 3), (5, 3), (5, 3)])

        assert isinstance(
            result.mock_class.one_tuple.unique_prior_tuples[0][1], af.GaussianPrior
        )
        assert isinstance(
            result.mock_class.one_tuple.unique_prior_tuples[1][1], af.GaussianPrior
        )
        assert isinstance(result.mock_class.two, af.GaussianPrior)

    def test_replace_priors_for_two_classes(self):
        mapper = af.ModelMapper(one=MockClassx2, two=MockClassx2)

        result = mapper.mapper_from_gaussian_tuples([(1, 1), (2, 1), (3, 1), (4, 1)])

        assert result.one.one.mean == 1
        assert result.one.two.mean == 2
        assert result.two.one.mean == 3
        assert result.two.two.mean == 4


class TestArguments:
    def test_same_argument_name(self):
        mapper = af.ModelMapper()

        mapper.one = af.PriorModel(MockClassx2)
        mapper.two = af.PriorModel(MockClassx2)

        instance = mapper.instance_from_vector([0.1, 0.2, 0.3, 0.4])

        assert instance.one.one == 0.1
        assert instance.one.two == 0.2
        assert instance.two.one == 0.3
        assert instance.two.two == 0.4


class TestIndependentPriorModel:
    def test_associate_prior_model(self):
        prior_model = af.PriorModel(MockClassx2)

        mapper = af.ModelMapper()

        mapper.prior_model = prior_model

        assert len(mapper.prior_model_tuples) == 1

        instance = mapper.instance_from_vector([0.1, 0.2])

        assert instance.prior_model.one == 0.1
        assert instance.prior_model.two == 0.2


@pytest.fixture(name="list_prior_model")
def make_list_prior_model():
    return af.CollectionPriorModel(
        [af.PriorModel(MockClassx2), af.PriorModel(MockClassx2)]
    )


class TestListPriorModel:
    def test_instance_from_vector(self, list_prior_model):
        mapper = af.ModelMapper()
        mapper.list = list_prior_model

        instance = mapper.instance_from_vector([0.1, 0.2, 0.3, 0.4])

        assert isinstance(instance.list, af.ModelInstance)
        print(instance.list.items)
        assert len(instance.list) == 2
        assert instance.list[0].one == 0.1
        assert instance.list[0].two == 0.2
        assert instance.list[1].one == 0.3
        assert instance.list[1].two == 0.4

    def test_prior_results_for_gaussian_tuples(self, list_prior_model):
        mapper = af.ModelMapper()
        mapper.list = list_prior_model

        gaussian_mapper = mapper.mapper_from_gaussian_tuples(
            [(1, 5), (2, 5), (3, 5), (4, 5)]
        )

        assert len(gaussian_mapper.list) == 2
        assert gaussian_mapper.list[0].one.mean == 1
        assert gaussian_mapper.list[0].two.mean == 2
        assert gaussian_mapper.list[1].one.mean == 3
        assert gaussian_mapper.list[1].two.mean == 4
        assert gaussian_mapper.list[0].one.sigma == 5
        assert gaussian_mapper.list[0].two.sigma == 5
        assert gaussian_mapper.list[1].one.sigma == 5
        assert gaussian_mapper.list[1].two.sigma == 5

    def test_prior_results_for_gaussian_tuples__include_override_from_width_file(
            self, list_prior_model
    ):
        mapper = af.ModelMapper()
        mapper.list = list_prior_model

        gaussian_mapper = mapper.mapper_from_gaussian_tuples(
            [(1, 0), (2, 0), (3, 0), (4, 0)]
        )

        assert len(gaussian_mapper.list) == 2
        assert gaussian_mapper.list[0].one.mean == 1
        assert gaussian_mapper.list[0].two.mean == 2
        assert gaussian_mapper.list[1].one.mean == 3
        assert gaussian_mapper.list[1].two.mean == 4
        assert gaussian_mapper.list[0].one.sigma == 1
        assert gaussian_mapper.list[0].two.sigma == 2
        assert gaussian_mapper.list[1].one.sigma == 1
        assert gaussian_mapper.list[1].two.sigma == 2

    def test_automatic_boxing(self):
        mapper = af.ModelMapper()
        mapper.list = [af.PriorModel(MockClassx2), af.PriorModel(MockClassx2)]

        assert isinstance(mapper.list, af.CollectionPriorModel)


@pytest.fixture(name="mock_with_instance")
def make_mock_with_instance():
    mock_with_instance = af.PriorModel(MockClassx2)
    mock_with_instance.one = 3.0
    return mock_with_instance


class Testinstance:
    def test__instance_prior_count(self, mock_with_instance):
        mapper = af.ModelMapper()
        mapper.mock_class = mock_with_instance

        assert len(mapper.unique_prior_tuples) == 1

    def test__retrieve_instances(self, mock_with_instance):
        assert len(mock_with_instance.instance_tuples) == 1

    def test_instance_prior_reconstruction(self, mock_with_instance):
        mapper = af.ModelMapper()
        mapper.mock_class = mock_with_instance

        instance = mapper.instance_for_arguments({mock_with_instance.two: 0.5})

        assert instance.mock_class.one == 3
        assert instance.mock_class.two == 0.5

    def test__instance_in_config(self):
        mapper = af.ModelMapper()

        mock_with_instance = af.PriorModel(MockClassx2instance, one=3)

        mapper.mock_class = mock_with_instance

        instance = mapper.instance_for_arguments({mock_with_instance.two: 0.5})

        assert instance.mock_class.one == 3
        assert instance.mock_class.two == 0.5

    def test__set_float(self):
        prior_model = af.PriorModel(MockClassx2)
        prior_model.one = 3
        prior_model.two = 4.0
        assert prior_model.one == 3
        assert prior_model.two == 4.0

    def test__list_prior_model_instances(self, mapper):
        prior_model = af.PriorModel(MockClassx2)
        prior_model.one = 3.0
        prior_model.two = 4.0

        mapper.mock_list = [prior_model]
        assert isinstance(mapper.mock_list, af.CollectionPriorModel)
        assert len(mapper.instance_tuples) == 2

    def test__set_for_tuple_prior(self):
        prior_model = af.PriorModel(MockChildTuplex3)
        prior_model.tup_0 = 1.0
        prior_model.tup_1 = 2.0
        prior_model.one = 1.0
        prior_model.two = 1.0
        prior_model.three = 1.0
        instance = prior_model.instance_for_arguments({})
        assert instance.tup == (1.0, 2.0)


@pytest.fixture(name="mock_config")
def make_mock_config():
    return


@pytest.fixture(name="mapper_with_one")
def make_mapper_with_one():
    mapper = af.ModelMapper()
    mapper.one = af.PriorModel(MockClassx2)
    return mapper


@pytest.fixture(name="mapper_with_list")
def make_mapper_with_list():
    mapper = af.ModelMapper()
    mapper.list = [af.PriorModel(MockClassx2), af.PriorModel(MockClassx2)]
    return mapper


class TestGaussianWidthConfig:
    def test_relative_widths(self, mapper):
        mapper.relative_width = MockClassRelativeWidth
        new_mapper = mapper.mapper_from_gaussian_tuples([(1, 0), (1, 0), (1, 0)])

        assert new_mapper.relative_width.one.mean == 1.0
        assert new_mapper.relative_width.one.sigma == 0.1

        assert new_mapper.relative_width.two.mean == 1.0
        assert new_mapper.relative_width.two.sigma == 0.5

        assert new_mapper.relative_width.three.mean == 1.0
        assert new_mapper.relative_width.three.sigma == 1.0

    def test_prior_classes(self, mapper_with_one):
        assert mapper_with_one.prior_class_dict == {
            mapper_with_one.one.one: MockClassx2,
            mapper_with_one.one.two: MockClassx2,
        }

    def test_prior_classes_list(self, mapper_with_list):
        assert mapper_with_list.prior_class_dict == {
            mapper_with_list.list[0].one: MockClassx2,
            mapper_with_list.list[0].two: MockClassx2,
            mapper_with_list.list[1].one: MockClassx2,
            mapper_with_list.list[1].two: MockClassx2,
        }

    def test_no_override(self):
        mapper = af.ModelMapper()

        mapper.one = af.PriorModel(MockClassx2)

        af.ModelMapper()

        assert mapper.one is not None



import pytest

import autofit as af
from autofit import graphical as g


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.GaussianPrior(100, 10)


@pytest.fixture(
    name="hierarchical_factor"
)
def make_hierarchical_factor(prior):
    factor = g.HierarchicalFactor(
        af.GaussianPrior,
        mean=af.GaussianPrior(
            mean=100,
            sigma=10
        ),
        sigma=af.GaussianPrior(
            mean=10,
            sigma=1
        )
    )
    factor.add_drawn_variable(
        prior
    )
    return factor



import pytest

import autofit as af


@pytest.fixture(
    name="model"
)
def make_model():
    return af.Model(
        af.Gaussian,
        centre=af.UniformPrior(
            lower_limit=0.0,
            upper_limit=1.0
        )
    )


@pytest.fixture(
    name="result"
)
def make_result(model):
    return af.GridSearchResult(
        results=[],
        lower_limits_lists=[
            [0.0],
            [0.5]
        ],
        grid_priors=[model.centre]
    )


@pytest.mark.parametrize(
    "upper_limit, physical_value",
    [
        (1.0, 0.5),
        (2.0, 1.0),
        (4.0, 2.0),
    ]
)
def test_physical_lower_limits(
        upper_limit,
        physical_value,
        model,
        result
):
    result.grid_priors = [
        af.UniformPrior(
            lower_limit=0.0,
            upper_limit=upper_limit
        )
    ]
    assert result.physical_lower_limits_lists == [
        [0.0],
        [physical_value]
    ]


def test_limits_lists(result):
    assert result.lower_limits_lists == [
        [0.0], [0.5]
    ]
    assert result.upper_limits_lists == [
        [0.5], [1.0]
    ]


def test_physical_centres_lists(
        result
):
    assert result.physical_centres_lists == [
        [0.25], [0.75]
    ]


def test_physical_upper_limits_lists(
        result
):
    assert result.physical_upper_limits_lists == [
        [0.5], [1.0]
    ]


def test_log_uniform_prior(
        result
):
    result.grid_priors = [af.LogUniformPrior()]

    assert result.physical_lower_limits_lists == [[1e-06], [pytest.approx(0.001, rel=0.01)]]
    assert result.physical_centres_lists == [
        [pytest.approx(3.1622776601683795e-05, rel=0.01)], [pytest.approx(0.03162277660168379, rel=0.01)]
    ]
    assert result.physical_upper_limits_lists == [[pytest.approx(0.001, rel=0.01)], [pytest.approx(1.0, rel=0.01)]]



from random import random

import dill
import numpy as np
import pytest

import autofit as af
from autofit import graphical as g
from autofit.messages.normal import UniformNormalMessage
from autofit.messages.transform import log_10_transform


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.GaussianPrior(
        mean=1,
        sigma=2
    )


def test():
    mean_field = g.MeanField({

    })
    mean_field.instance_for_arguments({})


def test_retain_id(
        prior
):
    new_message = prior * prior
    assert new_message.id == prior.id


@pytest.fixture(
    name="x"
)
def make_x():
    return np.linspace(
        0, 1, 100
    )


def test_projected_model():
    model = af.Model(
        af.Gaussian,
        centre=af.UniformPrior()
    )
    samples = af.Samples(
        model,
        [
            af.Sample(
                -1.0, -1.0,
                weight=random(),
                kwargs={
                    ("centre",): random(),
                    ("normalization",): random(),
                    ("sigma",): random(),
                }
            )
            for _ in range(100)
        ]
    )
    result = af.Result(
        samples=samples,
        model=model
    )
    projected_model = result.projected_model

    assert projected_model.prior_count == 3
    assert projected_model.centre is not model.centre
    assert projected_model.centre.id == model.centre.id
    assert isinstance(
        projected_model.centre,
        af.UniformPrior
    )


def test_uniform_normal(x):
    message = UniformNormalMessage.shifted(
        shift=1,
        scale=2.1
    )(
        mean=0.0,
        sigma=1.0
    )

    assert message.pdf(0.9) == 0
    assert message.pdf(3.2) == 0
    assert message.pdf(1.5) > 0


@pytest.mark.parametrize(
    "lower_limit, upper_limit, unit_value, physical_value",
    [
        (0.0, 1.0, 0.5, 0.5),
        (0.0, 1.0, 1.0, 1.0),
        (0.0, 1.0, 0.0, 0.0),
        (1.0, 2.0, 0.5, 1.5),
        (1.0, 2.0, 1.0, 2.0),
        (1.0, 2.0, 0.0, 1.0),
        (0.0, 2.0, 0.5, 1.0),
        (0.0, 2.0, 1.0, 2.0),
        (0.0, 2.0, 0.0, 0.0),
    ]
)
def test_uniform_prior(
        lower_limit,
        upper_limit,
        unit_value,
        physical_value
):
    assert af.UniformPrior(
        lower_limit=lower_limit,
        upper_limit=upper_limit,
    ).value_for(
        unit_value
    ) == pytest.approx(
        physical_value
    )


def test_uniform_odd_result():
    prior = af.UniformPrior(90.0, 100.0)
    assert prior.value_for(
        0.0
    ) == pytest.approx(90.0)


@pytest.mark.parametrize(
    "lower_limit",
    [
        1, 90
    ]
)
@pytest.mark.parametrize(
    "upper_limit",
    [
        110, 200
    ]
)
@pytest.mark.parametrize(
    "unit",
    [
        0.00001, 0.5, 0.9
    ]
)
def test_log10(
        lower_limit,
        upper_limit,
        unit
):
    prior = af.LogUniformPrior(
        lower_limit=lower_limit,
        upper_limit=upper_limit
    )

    assert 10.0 ** (
            np.log10(lower_limit)
            + unit * (np.log10(upper_limit) - np.log10(lower_limit))
    ) == pytest.approx(
        prior.value_for(
            unit
        ),
        abs=0.001
    )


@pytest.fixture(
    name="uniform_prior"
)
def make_uniform_prior():
    return af.UniformPrior(
        lower_limit=10,
        upper_limit=20,
        id_=1
    )


def test_prior_arithmetic(
        uniform_prior
):
    multiplied = uniform_prior * uniform_prior
    divided = multiplied / uniform_prior

    multiplied_value = multiplied.value_for(0.3)
    divided_value = divided.value_for(0.3)
    uniform_prior_value = uniform_prior.value_for(0.3)

    assert multiplied_value != divided_value
    assert divided_value == uniform_prior_value


def test_pickle_uniform_prior(
        uniform_prior
):
    pickled_prior = dill.loads(
        dill.dumps(uniform_prior)
    )
    assert pickled_prior == uniform_prior
    assert pickled_prior.id == uniform_prior.id


def test_pickle_log_uniform_prior():
    log_uniform_prior = af.LogUniformPrior()
    pickled_prior = dill.loads(
        dill.dumps(log_uniform_prior)
    )
    assert pickled_prior == log_uniform_prior


@pytest.fixture(
    name="LogMessage"
)
def make_log_message():
    return UniformNormalMessage.shifted(
        shift=1,
        scale=2,
    ).transformed(
        log_10_transform
    )


def test_pickle_transformed(
        LogMessage
):
    dill.loads(
        dill.dumps(LogMessage)
    )


def test_pickle_transformed_instantiated(
        LogMessage
):
    instance = LogMessage(
        mean=1,
        sigma=2
    )
    dill.loads(
        dill.dumps(instance)
    )



import pytest

import autofit as af

from autofit.mock.mock_model import MockClassx4
from autofit.mock.mock import MockNestSamples

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")

def test__samples_within_parameter_range(samples_x5):
    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    samples_x5 = MockNestSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
        total_samples=10,
        log_evidence=0.0,
        number_live_points=5,
    )

    samples_range = samples_x5.samples_within_parameter_range(
        parameter_index=0, parameter_range=[-1.0, 100.0]
    )

    assert len(samples_range.parameter_lists) == 5
    assert samples_x5.parameter_lists[0] == samples_range.parameter_lists[0]

    samples_range = samples_x5.samples_within_parameter_range(
        parameter_index=0, parameter_range=[1.0, 100.0]
    )

    assert len(samples_range.parameter_lists) == 1
    assert samples_range.parameter_lists[0] == [21.0, 22.0, 23.0, 24.0]

    samples_range = samples_x5.samples_within_parameter_range(
        parameter_index=2, parameter_range=[1.5, 2.5]
    )

    assert len(samples_range.parameter_lists) == 4
    assert samples_range.parameter_lists[0] == [0.0, 1.0, 2.0, 3.0]
    assert samples_range.parameter_lists[1] == [0.0, 1.0, 2.0, 3.0]
    assert samples_range.parameter_lists[2] == [0.0, 1.0, 2.0, 3.0]
    assert samples_range.parameter_lists[3] == [0.0, 1.0, 2.0, 3.0]


def test__acceptance_ratio_is_correct():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    samples_x5 = MockNestSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=5 * [[]],
            log_likelihood_list=[1.0, 2.0, 3.0, 4.0, 5.0],
            log_prior_list=5 * [0.0],
            weight_list=5 * [0.0],
        ),
        total_samples=10,
        log_evidence=0.0,
        number_live_points=5,
    )

    assert samples_x5.acceptance_ratio == 0.5



import pytest
import autofit as af

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class TestZeusConfig:

    def test__loads_from_config_file_correct(self):

        zeus = af.Zeus(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            nwalkers=51,
            nsteps=2001,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            auto_correlations_settings=af.AutoCorrelationsSettings(
                check_for_convergence=False,
                check_size=101,
                required_length=51,
                change_threshold=0.02
            ),
            tune=False,
            number_of_cores=2,
        )

        assert zeus.prior_passer.sigma == 2.0
        assert zeus.prior_passer.use_errors is False
        assert zeus.prior_passer.use_widths is False
        assert zeus.config_dict_search["nwalkers"] == 51
        assert zeus.config_dict_run["nsteps"] == 2001
        assert zeus.config_dict_run["tune"] == False
        assert isinstance(zeus.initializer, af.InitializerBall)
        assert zeus.initializer.lower_limit == 0.2
        assert zeus.initializer.upper_limit == 0.8
        assert zeus.auto_correlations_settings.check_for_convergence is False
        assert zeus.auto_correlations_settings.check_size == 101
        assert zeus.auto_correlations_settings.required_length == 51
        assert zeus.auto_correlations_settings.change_threshold == 0.02
        assert zeus.number_of_cores == 2

        zeus = af.Zeus()

        assert zeus.prior_passer.sigma == 3.0
        assert zeus.prior_passer.use_errors is True
        assert zeus.prior_passer.use_widths is True
        assert zeus.config_dict_search["nwalkers"] == 50
        assert zeus.config_dict_run["nsteps"] == 2000
        assert zeus.config_dict_run["tune"] == True
        assert isinstance(zeus.initializer, af.InitializerPrior)
        assert zeus.auto_correlations_settings.check_for_convergence is True
        assert zeus.auto_correlations_settings.check_size == 100
        assert zeus.auto_correlations_settings.required_length == 50
        assert zeus.auto_correlations_settings.change_threshold == 0.01
        assert zeus.number_of_cores == 1




import pytest

import autofit as af
from autofit import database as db
from autofit.non_linear.samples import Samples


@pytest.fixture(
    name="model"
)
def make_model():
    return af.PriorModel(
        af.Gaussian
    )


@pytest.fixture(
    name="serialized_model"
)
def make_serialized_model(model):
    return db.Object.from_object(
        model
    )


@pytest.fixture(
    name="collection"
)
def make_collection(model):
    return af.CollectionPriorModel(
        model=model
    )


@pytest.fixture(
    name="serialized_collection"
)
def make_serialized_collection(collection):
    return db.Object.from_object(
        collection
    )


class TestInstance:
    def test_serialize(self):
        serialized_instance = db.Object.from_object(
            af.Gaussian()
        )
        assert len(
            serialized_instance.children
        ) == 3

    def test_tuple(self):
        assert db.Object.from_object(
            (1, 2)
        )() == (1, 2)


class TestModel:
    def test_serialize(
            self,
            serialized_model
    ):
        assert isinstance(
            serialized_model, db.PriorModel
        )
        assert serialized_model.cls is af.Gaussian

    def test_deserialize(
            self,
            serialized_model
    ):
        assert serialized_model().cls is af.Gaussian


class TestPriors:
    def test_serialize(
            self,
            serialized_model,
    ):
        assert len(serialized_model.children) == 3

    def test_deserialize(
            self,
            serialized_model
    ):
        model = serialized_model()
        assert len(model.priors) == 3
        assert isinstance(
            model.centre,
            af.UniformPrior
        )


class TestCollection:
    def test_serialize(
            self,
            serialized_collection
    ):
        assert isinstance(
            serialized_collection,
            db.CollectionPriorModel
        )
        child, = serialized_collection.children
        assert len(child.children) == 3

    def test_deserialize(
            self,
            serialized_collection
    ):
        collection = serialized_collection()
        assert len(collection) == 1
        assert isinstance(
            collection.model,
            af.PriorModel
        )


class TestClasses:
    def test_samples(self):
        db.Object.from_object(
            Samples(
                af.ModelMapper(),
                None
            )
        )()

    def test_string(self):
        assert "string" == db.Object.from_object(
            "string"
        )()


class Serialisable:
    def __init__(self):
        self.value = 1

    def __getstate__(self):
        return {
            "value": 2 * self.value
        }

    def __setstate__(self, state):
        state["value"] *= -1
        self.__dict__.update(
            state
        )


def test_get_set_state():
    assert db.Object.from_object(
        Serialisable()
    )().value == -2


def test_none():
    assert db.Object.from_object(None)() is None


def test_commit(session):
    model = af.PriorModel(
        af.Gaussian
    )
    serialized = db.Object.from_object(model)
    session.add(serialized)
    session.commit()



import numpy as np
import pytest

from autoconf.conf import with_config

import autofit as af
from autofit.mock.mock_model import MockClassx2, MockClassx4
from autofit.mock.mock import MockSamples

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


def test__from_csv_table():

    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    samples_x5 = af.Samples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )

    filename = "samples.csv"
    samples_x5.write_table(filename=filename)

    samples_x5 = af.PDFSamples.from_table(filename=filename, model=samples_x5.model)

    assert samples_x5.parameter_lists == [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]
    assert samples_x5.log_likelihood_list == [1.0, 2.0, 3.0, 10.0, 5.0]
    assert samples_x5.log_prior_list == [0.0, 0.0, 0.0, 0.0, 0.0]
    assert samples_x5.log_posterior_list == [1.0, 2.0, 3.0, 10.0, 5.0]
    assert samples_x5.weight_list == [1.0, 1.0, 1.0, 1.0, 1.0]


def test__converged__median_pdf_vector_and_instance():
    parameters = [
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [0.9, 1.9],
        [1.1, 2.1],
    ]

    log_likelihood_list = 10 * [0.1]
    weight_list = 10 * [0.1]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.pdf_converged is True

    median_pdf_vector = samples_x5.median_pdf_vector

    assert median_pdf_vector[0] == pytest.approx(1.0, 1.0e-4)
    assert median_pdf_vector[1] == pytest.approx(2.0, 1.0e-4)

    median_pdf_instance = samples_x5.median_pdf_instance

    assert median_pdf_instance.mock_class.one == pytest.approx(1.0, 1e-1)
    assert median_pdf_instance.mock_class.two == pytest.approx(2.0, 1e-1)


def test__unconverged__median_pdf_vector():
    parameters = [
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.1, 2.1],
        [0.9, 1.9],
    ]

    log_likelihood_list = 9 * [0.0] + [1.0]
    weight_list = 9 * [0.0] + [1.0]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.pdf_converged is False

    median_pdf_vector = samples_x5.median_pdf_vector

    assert median_pdf_vector[0] == pytest.approx(0.9, 1.0e-4)
    assert median_pdf_vector[1] == pytest.approx(1.9, 1.0e-4)


@with_config("general", "model", "ignore_prior_limits", value=True)
def test__converged__vector_and_instance_at_upper_and_lower_sigma():
    parameters = [
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.0, 0.5],
        [0.2, 0.3],
    ]

    log_likelihood_list = list(range(10))

    weight_list = 10 * [0.1]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.pdf_converged is True

    vector_at_sigma = samples_x5.vector_at_sigma(sigma=3.0)

    assert vector_at_sigma[0] == pytest.approx((0.00121, 0.19878), 1e-1)
    assert vector_at_sigma[1] == pytest.approx((0.30121, 0.49878), 1e-1)

    vector_at_sigma = samples_x5.vector_at_upper_sigma(sigma=3.0)

    assert vector_at_sigma[0] == pytest.approx(0.19757, 1e-1)
    assert vector_at_sigma[1] == pytest.approx(0.49757, 1e-1)

    vector_at_sigma = samples_x5.vector_at_lower_sigma(sigma=3.0)

    assert vector_at_sigma[0] == pytest.approx(0.00121, 1e-1)
    assert vector_at_sigma[1] == pytest.approx(0.30121, 1e-1)

    vector_at_sigma = samples_x5.vector_at_sigma(sigma=1.0)

    assert vector_at_sigma[0] == pytest.approx((0.1, 0.1), 1e-1)
    assert vector_at_sigma[1] == pytest.approx((0.4, 0.4), 1e-1)

    instance_at_sigma = samples_x5.instance_at_sigma(sigma=1.0)

    assert instance_at_sigma.mock_class.one == pytest.approx((0.1, 0.1), 1e-1)
    assert instance_at_sigma.mock_class.two == pytest.approx((0.4, 0.4), 1e-1)

    instance_at_sigma = samples_x5.instance_at_upper_sigma(sigma=3.0)

    assert instance_at_sigma.mock_class.one == pytest.approx(0.19757, 1e-1)
    assert instance_at_sigma.mock_class.two == pytest.approx(0.49757, 1e-1)

    instance_at_sigma = samples_x5.instance_at_lower_sigma(sigma=3.0)

    assert instance_at_sigma.mock_class.one == pytest.approx(0.00121, 1e-1)
    assert instance_at_sigma.mock_class.two == pytest.approx(0.30121, 1e-1)


def test__unconverged_vector_at_lower_and_upper_sigma():
    parameters = [
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.0, 2.0],
        [1.1, 2.1],
        [0.9, 1.9],
    ]

    log_likelihood_list = 9 * [0.0] + [1.0]
    weight_list = 9 * [0.0] + [1.0]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.pdf_converged is False

    vector_at_sigma = samples_x5.vector_at_sigma(sigma=1.0)

    assert vector_at_sigma[0] == pytest.approx(((0.9, 1.1)), 1e-2)
    assert vector_at_sigma[1] == pytest.approx(((1.9, 2.1)), 1e-2)

    vector_at_sigma = samples_x5.vector_at_sigma(sigma=3.0)

    assert vector_at_sigma[0] == pytest.approx(((0.9, 1.1)), 1e-2)
    assert vector_at_sigma[1] == pytest.approx(((1.9, 2.1)), 1e-2)


def test__converged__errors_vector_and_instance_at_upper_and_lower_sigma():
    parameters = [
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.1, 0.4],
        [0.0, 0.5],
        [0.2, 0.3],
    ]

    log_likelihood_list = list(range(10))

    weight_list = 10 * [0.1]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.pdf_converged is True

    errors = samples_x5.error_magnitude_vector_at_sigma(sigma=3.0)

    assert errors == pytest.approx([0.19514, 0.19514], 1e-1)

    errors = samples_x5.error_vector_at_upper_sigma(sigma=3.0)

    assert errors == pytest.approx([0.09757, 0.09757], 1e-1)

    errors = samples_x5.error_vector_at_lower_sigma(sigma=3.0)

    assert errors == pytest.approx([0.09757, 0.09757], 1e-1)

    errors = samples_x5.error_vector_at_sigma(sigma=3.0)
    assert errors[0] == pytest.approx((0.09757, 0.09757), 1e-1)
    assert errors[1] == pytest.approx((0.09757, 0.09757), 1e-1)

    errors = samples_x5.error_magnitude_vector_at_sigma(sigma=1.0)

    assert errors == pytest.approx([0.0, 0.0], 1e-1)

    errors_instance = samples_x5.error_instance_at_sigma(sigma=1.0)

    assert errors_instance.mock_class.one == pytest.approx(0.0, 1e-1)
    assert errors_instance.mock_class.two == pytest.approx(0.0, 1e-1)

    errors_instance = samples_x5.error_instance_at_upper_sigma(sigma=3.0)

    assert errors_instance.mock_class.one == pytest.approx(0.09757, 1e-1)
    assert errors_instance.mock_class.two == pytest.approx(0.09757, 1e-1)

    errors_instance = samples_x5.error_instance_at_lower_sigma(sigma=3.0)

    assert errors_instance.mock_class.one == pytest.approx(0.09757, 1e-1)
    assert errors_instance.mock_class.two == pytest.approx(0.09757, 1e-1)


def test__unconverged_sample_size__uses_value_unless_fewer_samples():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    log_likelihood_list = 4 * [0.0] + [1.0]
    weight_list = 4 * [0.0] + [1.0]

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=5 * [[]],
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[1.0, 1.0, 1.0, 1.0, 1.0],
            weight_list=weight_list,
        ),
        unconverged_sample_size=2,
    )

    assert samples_x5.pdf_converged is False
    assert samples_x5.unconverged_sample_size == 2

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=5 * [[]],
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[1.0, 1.0, 1.0, 1.0, 1.0],
            weight_list=weight_list,
        ),
        unconverged_sample_size=6,
    )

    assert samples_x5.pdf_converged is False
    assert samples_x5.unconverged_sample_size == 5


def test__offset_vector_from_input_vector():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [1.1, 2.1, 3.1, 4.1],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.1],
    ]

    weight_list = [0.3, 0.2, 0.2, 0.2, 0.1]

    log_likelihood_list = list(map(lambda weight: 10.0 * weight, weight_list))

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    offset_values = samples_x5.offset_vector_from_input_vector(
        input_vector=[1.0, 1.0, 2.0, 3.0]
    )

    assert offset_values == pytest.approx([0.0, 1.0, 1.0, 1.025], 1.0e-4)


def test__vector_drawn_randomly_from_pdf():
    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    model = af.ModelMapper(mock_class_1=MockClassx4)

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 4.0, 5.0],
            log_prior_list=5 * [0.0],
            weight_list=[0.0, 0.0, 0.0, 1.0, 0.0],
        ),
    )

    vector = samples_x5.vector_drawn_randomly_from_pdf()

    assert vector == [21.0, 22.0, 23.0, 24.0]

    instance = samples_x5.instance_drawn_randomly_from_pdf()

    assert vector == [21.0, 22.0, 23.0, 24.0]

    assert instance.mock_class_1.one == 21.0
    assert instance.mock_class_1.two == 22.0
    assert instance.mock_class_1.three == 23.0
    assert instance.mock_class_1.four == 24.0


def test__covariance_matrix():
    log_likelihood_list = list(range(3))

    weight_list = 3 * [0.1]

    parameters = [[2.0, 2.0], [1.0, 1.0], [0.0, 0.0]]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=3 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.covariance_matrix() == pytest.approx(
        np.array([[1.0, 1.0], [1.0, 1.0]]), 1.0e-4
    )

    parameters = [[0.0, 2.0], [1.0, 1.0], [2.0, 0.0]]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=3 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.covariance_matrix() == pytest.approx(
        np.array([[1.0, -1.0], [-1.0, 1.0]]), 1.0e-4
    )

    weight_list = [0.1, 0.2, 0.3]

    model = af.ModelMapper(mock_class=MockClassx2)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=10 * [0.0],
            weight_list=weight_list,
        ),
    )

    assert samples_x5.covariance_matrix() == pytest.approx(
        np.array([[0.90909, -0.90909], [-0.90909, 0.90909]]), 1.0e-4
    )



import pytest

import autofit as af
from autofit.mock.mock_model import MockWithTuple


@pytest.fixture(
    name="sample"
)
def make_sample():
    return af.Sample(
        log_likelihood=0,
        log_prior=0,
        weight=0,
        kwargs={
            ("gaussian_1", "centre",): 1,
            ("gaussian_1", "normalization",): 2,
            ("gaussian_1", "sigma",): 3,
            ("gaussian_2", "centre",): 4,
            ("gaussian_2", "normalization",): 5,
            ("gaussian_2", "sigma",): 6,
        }
    )


@pytest.fixture(
    name="samples"
)
def make_samples(
        model,
        sample
):
    return af.Samples(
        model=model,
        sample_list=[sample],
    )


class TestWith:
    def test_trivial(
            self,
            sample
    ):
        with_paths = sample.with_paths([
            ("gaussian_1", "centre",)
        ])

        assert with_paths.kwargs == {
            ("gaussian_1", "centre",): 1
        }

    def test_subpath(
            self,
            sample
    ):
        with_paths = sample.with_paths([
            ("gaussian_1",)
        ])

        assert with_paths.kwargs == {
            ("gaussian_1", "centre",): 1,
            ("gaussian_1", "normalization",): 2,
            ("gaussian_1", "sigma",): 3,
        }

    def test_samples(
            self,
            samples
    ):
        with_paths = samples.with_paths([
            ("gaussian_1",)
        ])

        assert with_paths.sample_list[0].kwargs == {
            ("gaussian_1", "centre",): 1,
            ("gaussian_1", "normalization",): 2,
            ("gaussian_1", "sigma",): 3,
        }

        model = with_paths.model
        assert hasattr(
            model,
            "gaussian_1"
        )
        assert not hasattr(
            model,
            "gaussian_2"
        )


class TestWithout:
    def test_trivial(
            self,
            sample
    ):
        without_paths = sample.without_paths([
            ("gaussian_1", "centre",)
        ])

        assert without_paths.kwargs == {
            ("gaussian_1", "normalization",): 2,
            ("gaussian_1", "sigma",): 3,
            ("gaussian_2", "centre",): 4,
            ("gaussian_2", "normalization",): 5,
            ("gaussian_2", "sigma",): 6,
        }

    def test_subpath(
            self,
            sample
    ):
        without_paths = sample.without_paths([
            ("gaussian_1",)
        ])

        assert without_paths.kwargs == {
            ("gaussian_2", "centre",): 4,
            ("gaussian_2", "normalization",): 5,
            ("gaussian_2", "sigma",): 6,
        }

    def test_samples(
            self,
            samples,
    ):
        without_paths = samples.without_paths([
            ("gaussian_1",)
        ])

        assert without_paths.sample_list[0].kwargs == {
            ("gaussian_2", "centre",): 4,
            ("gaussian_2", "normalization",): 5,
            ("gaussian_2", "sigma",): 6,
        }

        model = without_paths.model
        assert not hasattr(
            model,
            "gaussian_1"
        )
        assert hasattr(
            model,
            "gaussian_2"
        )


def test_samples_lazy_attributes(
        samples
):
    paths = samples.paths
    names = samples.names

    without_paths = samples.without_paths([])

    assert without_paths.paths is not paths
    assert without_paths.names is not names


def test_tuples(samples):
    model = af.Collection(
        lens=af.Model(
            MockWithTuple
        )
    )
    samples = af.Samples(
        model=model,
        sample_list=[
            af.Sample(
                log_likelihood=0,
                log_prior=0,
                weight=0,
                kwargs={
                    path: i
                    for i, path
                    in enumerate(
                        model.paths
                    )
                }
            )
        ]
    )
    samples = samples.without_paths(
        [
            ("lens", "tup",),
        ]
    )
    assert len(samples.parameter_lists) == 1



import pytest

import autofit as af
from autofit.mock.mock import MockAnalysis
from test_autofit.non_linear.grid.test_optimizer_grid_search import MockOptimizer


def test_is_database_paths(session):
    optimizer = MockOptimizer(
        session=session
    )
    assert isinstance(
        optimizer.paths,
        af.DatabasePaths
    )
    # noinspection PyUnresolvedReferences
    assert optimizer.paths.save_all_samples is False


@pytest.mark.parametrize(
    "save_all_samples",
    [True, False]
)
def test_save_all_samples_boolean(
        session,
        save_all_samples
):
    optimizer = MockOptimizer(
        session=session,
        save_all_samples=save_all_samples
    )
    # noinspection PyUnresolvedReferences
    assert optimizer.paths.save_all_samples is save_all_samples


@pytest.mark.parametrize(
    "save_all_samples, n_samples",
    [
        (False, 1),
        (True, 2)
    ]
)
def test_save_all_samples(
        session,
        save_all_samples,
        n_samples
):
    analysis = MockAnalysis()
    model = af.Model(af.Gaussian)

    optimizer = MockOptimizer(
        session=session,
        save_all_samples=save_all_samples,
        sample_multiplier=2
    )

    optimizer.fit(
        model,
        analysis
    )

    fit = session.query(
        af.db.Fit
    ).one()

    assert len(fit.samples.sample_list) == n_samples


def test_unique_tag(session):
    analysis = MockAnalysis()
    model = af.Model(af.Gaussian)

    unique_tag = "unique"

    optimizer = MockOptimizer(
        session=session,
        unique_tag=unique_tag
    )

    assert optimizer.paths.unique_tag == unique_tag

    optimizer.fit(
        model,
        analysis
    )

    fit = session.query(
        af.db.Fit
    ).one()

    assert fit.unique_tag == unique_tag



import pickle
from typing import Tuple

import numpy as np
import pytest

import autofit as af
from autofit import exc, Result
from autofit.graphical import FactorApproximation
from autofit.graphical.utils import Status
from autofit.mock.mock import MockAnalysis, MockSearch


def test_unpickle_result():
    # noinspection PyTypeChecker
    result = af.GridSearchResult(
        results=[af.Result(samples=None, model=None)],
        lower_limits_lists=[[1]],
        grid_priors=[],
    )
    result = pickle.loads(pickle.dumps(result))
    assert result is not None


class TestGridSearchablePriors:
    def test_generated_models(self, grid_search, mapper):

        mappers = list(
            grid_search.model_mappers(
                mapper,
                grid_priors=[
                    mapper.component.one_tuple.one_tuple_0,
                    mapper.component.one_tuple.one_tuple_1,
                ],
            )
        )

        assert len(mappers) == 100

        assert mappers[0].component.one_tuple.one_tuple_0.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_0.upper_limit == 0.1
        assert mappers[0].component.one_tuple.one_tuple_1.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_1.upper_limit == 0.2

        assert mappers[-1].component.one_tuple.one_tuple_0.lower_limit == 0.9
        assert mappers[-1].component.one_tuple.one_tuple_0.upper_limit == 1.0
        assert mappers[-1].component.one_tuple.one_tuple_1.lower_limit == 1.8
        assert mappers[-1].component.one_tuple.one_tuple_1.upper_limit == 2.0

    def test_non_grid_searched_dimensions(self, mapper):
        search = MockSearch()
        search.paths = af.DirectoryPaths(name="")
        grid_search = af.SearchGridSearch(
            number_of_steps=10,
            search=search
        )

        mappers = list(
            grid_search.model_mappers(
                mapper, grid_priors=[mapper.component.one_tuple.one_tuple_0]
            )
        )

        assert len(mappers) == 10

        assert mappers[0].component.one_tuple.one_tuple_0.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_0.upper_limit == 0.1
        assert mappers[0].component.one_tuple.one_tuple_1.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_1.upper_limit == 2.0

        assert mappers[-1].component.one_tuple.one_tuple_0.lower_limit == 0.9
        assert mappers[-1].component.one_tuple.one_tuple_0.upper_limit == 1.0
        assert mappers[-1].component.one_tuple.one_tuple_1.lower_limit == 0.0
        assert mappers[-1].component.one_tuple.one_tuple_1.upper_limit == 2.0

    def test_tied_priors(self, grid_search, mapper):
        mapper.component.one_tuple.one_tuple_0 = mapper.component.one_tuple.one_tuple_1

        mappers = list(
            grid_search.model_mappers(
                grid_priors=[
                    mapper.component.one_tuple.one_tuple_0,
                    mapper.component.one_tuple.one_tuple_1,
                ],
                model=mapper,
            )
        )

        assert len(mappers) == 10

        assert mappers[0].component.one_tuple.one_tuple_0.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_0.upper_limit == 0.2
        assert mappers[0].component.one_tuple.one_tuple_1.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_1.upper_limit == 0.2

        assert mappers[-1].component.one_tuple.one_tuple_0.lower_limit == 1.8
        assert mappers[-1].component.one_tuple.one_tuple_0.upper_limit == 2.0
        assert mappers[-1].component.one_tuple.one_tuple_1.lower_limit == 1.8
        assert mappers[-1].component.one_tuple.one_tuple_1.upper_limit == 2.0

        for mapper in mappers:
            assert (
                    mapper.component.one_tuple.one_tuple_0
                    == mapper.component.one_tuple.one_tuple_1
            )

    def test_different_prior_width(self, grid_search, mapper):
        mapper.component.one_tuple.one_tuple_0 = af.UniformPrior(0.0, 2.0)
        mappers = list(
            grid_search.model_mappers(
                grid_priors=[mapper.component.one_tuple.one_tuple_0], model=mapper
            )
        )

        assert len(mappers) == 10

        assert mappers[0].component.one_tuple.one_tuple_0.lower_limit == 0.0
        assert mappers[0].component.one_tuple.one_tuple_0.upper_limit == 0.2

        assert mappers[-1].component.one_tuple.one_tuple_0.lower_limit == 1.8
        assert mappers[-1].component.one_tuple.one_tuple_0.upper_limit == 2.0

        mapper.component.one_tuple.one_tuple_0 = af.UniformPrior(1.0, 1.5)
        mappers = list(
            grid_search.model_mappers(
                mapper, grid_priors=[mapper.component.one_tuple.one_tuple_0]
            )
        )

        assert len(mappers) == 10

        assert mappers[0].component.one_tuple.one_tuple_0.lower_limit == 1.0
        assert mappers[0].component.one_tuple.one_tuple_0.upper_limit == 1.05

        assert mappers[-1].component.one_tuple.one_tuple_0.lower_limit == 1.45
        assert mappers[-1].component.one_tuple.one_tuple_0.upper_limit == 1.5

    def test_raises_exception_for_bad_limits(self, grid_search, mapper):
        mapper.component.one_tuple.one_tuple_0 = af.GaussianPrior(
            0.0, 2.0, lower_limit=float("-inf"), upper_limit=float("inf")
        )
        with pytest.raises(exc.PriorException):
            list(
                grid_search.make_arguments(
                    [[0, 1]], grid_priors=[mapper.component.one_tuple.one_tuple_0]
                )
            )


@pytest.fixture(name="grid_search_05")
def make_grid_search_05():
    search = af.SearchGridSearch(
        search=MockOptimizer(), number_of_steps=2
    )
    search.search.paths = af.DirectoryPaths(name="sample_name")
    return search


class MockOptimizer(MockSearch):

    def __init__(self, **kwargs):
        super().__init__(fit_fast=False, **kwargs)

    @property
    def samples_cls(self):
        return MockOptimizer

    def project(
            self,
            factor_approx: FactorApproximation,
            status: Status = Status()
    ) -> Tuple[FactorApproximation, Status]:
        pass

    init_args = list()


@pytest.fixture(autouse=True)
def empty_args():
    MockOptimizer.init_args = list()


class TestGridNLOBehaviour:
    def test_results(self, grid_search_05, mapper):
        result = grid_search_05.fit(
            model=mapper,
            analysis=MockAnalysis(),
            grid_priors=[
                mapper.component.one_tuple.one_tuple_0,
                mapper.component.one_tuple.one_tuple_1,
            ],
        )

        assert len(result.results) == 4
        assert result.no_dimensions == 2

        grid_search = af.SearchGridSearch(
            search=MockOptimizer(),
            number_of_steps=10,
        )
        grid_search.search.paths = af.DirectoryPaths(name="sample_name")
        result = grid_search.fit(
            model=mapper,
            analysis=MockAnalysis(),
            grid_priors=[
                mapper.component.one_tuple.one_tuple_0,
                mapper.component.one_tuple.one_tuple_1,
            ],
        )

        assert len(result.results) == 100
        assert result.no_dimensions == 2
        assert result.log_likelihoods_native.shape == (10, 10)

    # def test_results_parallel(self, mapper, container):
    #     grid_search = af.SearchGridSearch(
    #         search=container.MockOptimizer,
    #         number_of_steps=10,
    #         paths=af.Paths(name="sample_name"),
    #         parallel=True,
    #     )
    #     result = grid_search.fit(
    #         container.MockAnalysis(),
    #         mapper,
    #         [mapper.component.one_tuple.one_tuple_0, mapper.component.one_tuple.one_tuple_1],
    #     )
    #
    #     assert len(result.results) == 100
    #     assert result.no_dimensions == 2
    #     assert result.likelihood_merit_array.shape == (10, 10)

    # def test_generated_models_with_instances(self, grid_search, container, mapper):
    #     instance_component = mock.MockClassx2Tuple()
    #     mapper.instance_component = instance_component
    #
    #     analysis = container.MockAnalysis()
    #
    #     grid_search.fit(analysis, mapper, [mapper.component.one_tuple.one_tuple_0])
    #
    #     for instance in container.fit_instances:
    #         assert isinstance(instance.component, mock.MockClassx2Tuple)
    #         assert instance.instance_component == instance_component
    #
    # def test_generated_models_with_instance_attributes(
    #         self, grid_search, mapper, container
    # ):
    #     instance = 2.0
    #     mapper.component.one_tuple.one_tuple_1 = instance
    #
    #     analysis = container.MockAnalysis()
    #
    #     grid_search.fit(analysis, mapper, [mapper.component.one_tuple.one_tuple_0])
    #
    #     assert len(container.fit_instances) > 0
    #
    #     for instance in container.fit_instances:
    #         assert isinstance(instance.component, mock.MockClassx2Tuple)
    #         # noinspection PyUnresolvedReferences
    #         assert instance.component.centre[1] == 2

    def test_passes_attributes(self):
        search = af.DynestyStatic()
        search.paths = af.DirectoryPaths(name="")
        grid_search = af.SearchGridSearch(
            number_of_steps=10,
            search=search
        )

        grid_search.nlive = 20
        grid_search.facc = 0.3

        search = grid_search.search_instance("name_path")

        assert search.nlive is grid_search.nlive
        assert grid_search.paths.path != search.paths.path
        assert grid_search.paths.output_path != search.paths.output_path


class MockResult(Result):
    def __init__(self, log_likelihood):
        # noinspection PyTypeChecker
        super().__init__(None, None)
        self._log_likelihood = log_likelihood
        self.model = log_likelihood

    @property
    def log_likelihood(self):
        return self._log_likelihood


@pytest.fixture(name="grid_search_result")
def make_grid_search_result():
    one = MockResult(1)
    two = MockResult(2)

    # noinspection PyTypeChecker
    return af.GridSearchResult(
        results=[one, two],
        lower_limits_lists=[[1], [2]],
        grid_priors=[[1], [2]]
    )


class TestGridSearchResult:

    def test_best_result(self, grid_search_result):
        assert grid_search_result.best_result.log_likelihood == 2

    def test_attributes(self, grid_search_result):
        assert grid_search_result.model == 2

    def test_best_model(self, grid_search_result):
        assert grid_search_result.best_model == 2

    def test_all_models(self, grid_search_result):
        assert grid_search_result.all_models == [1, 2]

    def test__result_derived_properties(self):
        lower_limit_lists = [[0.0, 0.0], [0.0, 0.5], [0.5, 0.0], [0.5, 0.5]]

        # noinspection PyTypeChecker
        grid_search_result = af.GridSearchResult(
            results=None,
            grid_priors=[
                af.UniformPrior(
                    lower_limit=-2.0,
                    upper_limit=2.0
                ),
                af.UniformPrior(
                    lower_limit=-3.0,
                    upper_limit=3.0
                )
            ],
            lower_limits_lists=lower_limit_lists,
        )

        assert grid_search_result.shape == (2, 2)
        assert grid_search_result.physical_step_sizes == (2.0, 3.0)
        assert grid_search_result.physical_centres_lists == [
            [-1.0, -1.5],
            [-1.0, 1.5],
            [1.0, -1.5],
            [1.0, 1.5],
        ]
        assert grid_search_result.physical_upper_limits_lists == [
            [0.0, 0.0],
            [0.0, 3.0],
            [2.0, 0.0],
            [2.0, 3.0],
        ]

    def test__results_on_native_grid(self, grid_search_result):
        assert (grid_search_result.results_native == np.array([
            [grid_search_result.results[0], grid_search_result.results[1]],
        ])).all()

        assert (grid_search_result.log_likelihoods_native == np.array([
            [1, 2],
        ])).all()


@pytest.mark.parametrize(
    "n_dimensions, n_steps",
    [
        (2, 2),
        (3, 3),
        (2, 3),
        (3, 2),
        (4, 4),
    ]
)
def test_higher_dimensions(
        n_dimensions,
        n_steps
):
    shape = n_dimensions * (n_steps,)
    total = n_steps ** n_dimensions
    model = af.Model(
        af.Gaussian
    )
    result = af.GridSearchResult(
        results=total * [
            af.Result(
                af.NestSamples(
                    model,
                    [
                        af.Sample(
                            1.0,
                            1.0,
                            1.0,
                            {
                                "centre": 1.0,
                                "sigma": 1.0,
                                "normalization": 1.0
                            }
                        )
                    ]
                ),
                model=model
            )
        ],
        grid_priors=[],
        lower_limits_lists=total * [
            n_dimensions * [0.0]
        ]
    )
    assert result.shape == shape
    assert result.results_native.shape == shape
    assert result.log_likelihoods_native.shape == shape



import datetime as dt


def test_import():
    """
    Ensure that import time is less than 100 microseconds
    """
    tic = dt.datetime.now()
    # noinspection PyUnresolvedReferences
    import autofit
    difference = (dt.datetime.now() - tic).microseconds
    assert difference < 100



import pytest

import autofit as af
from autofit.mapper.prior_model.abstract import paths_to_tree
from autofit.mock.mock_model import MockWithTuple


class TestWithoutAttributes:
    @pytest.mark.parametrize(
        "attribute",
        [
            "centre",
            "normalization",
            "sigma",
        ]
    )
    def test_gaussian(
            self,
            gaussian_1,
            attribute
    ):
        without = gaussian_1.without_attributes()
        assert not hasattr(
            without,
            attribute
        )
        assert hasattr(
            without, "cls"
        )

    @pytest.mark.parametrize(
        "attribute",
        [
            "gaussian_1",
            "gaussian_2"
        ]
    )
    def test_collection(
            self,
            model,
            attribute
    ):
        without = model.without_attributes()
        assert not hasattr(
            without,
            attribute
        )


class TestWith:
    def test_subpath(self, model):
        with_paths = model.with_paths([
            ("gaussian_1",)
        ])
        assert not hasattr(
            with_paths,
            "gaussian_2"
        )
        assert hasattr(
            with_paths.gaussian_1,
            "centre"
        )

    def test_path(self, model):
        with_paths = model.with_paths([
            ("gaussian_1", "centre"),
            ("gaussian_2", "normalization"),
        ])

        gaussian_1 = with_paths.gaussian_1
        assert hasattr(gaussian_1, "centre")
        assert not hasattr(gaussian_1, "normalization")

        gaussian_2 = with_paths.gaussian_2
        assert not hasattr(gaussian_2, "centre")
        assert hasattr(gaussian_2, "normalization")


def test_string_paths(model):
    with_paths = model.with_paths([
        "gaussian_1.centre",
        "gaussian_2.normalization",
    ])

    gaussian_1 = with_paths.gaussian_1
    assert hasattr(gaussian_1, "centre")
    assert not hasattr(gaussian_1, "normalization")

    gaussian_2 = with_paths.gaussian_2
    assert not hasattr(gaussian_2, "centre")
    assert hasattr(gaussian_2, "normalization")


class TestWithout:
    def test_subpath(self, model):
        with_paths = model.without_paths([
            ("gaussian_1",)
        ])
        assert not hasattr(
            with_paths,
            "gaussian_1"
        )
        assert hasattr(
            with_paths.gaussian_2,
            "centre"
        )

    def test_path(self, model):
        with_paths = model.without_paths([
            ("gaussian_1", "centre"),
            ("gaussian_2", "normalization"),
        ])

        gaussian_1 = with_paths.gaussian_1
        assert not hasattr(gaussian_1, "centre")
        assert hasattr(gaussian_1, "normalization")

        gaussian_2 = with_paths.gaussian_2
        assert hasattr(gaussian_2, "centre")
        assert not hasattr(gaussian_2, "normalization")


class TestPathsToTree:
    def test_trivial(self):
        assert paths_to_tree([
            ("one", "two")
        ]) == {
                   "one": {
                       "two": {}
                   }
               }

    def test_multiple(self):
        assert paths_to_tree([
            ("one", "two"),
            ("one", "three"),
            ("two", "three"),
        ]) == {
                   "one": {
                       "two": {},
                       "three": {},
                   },
                   "two": {
                       "three": {}
                   }
               }

    def test_second_tier(self):
        assert paths_to_tree([
            ("one", "two", "three"),
            ("one", "two", "four"),
        ]) == {
                   "one": {
                       "two": {
                           "three": {},
                           "four": {}
                       }
                   }
               }


@pytest.mark.parametrize(
    'path, index',
    [
        (("gaussian_1", "centre"), 0),
        (("gaussian_2", "centre"), 3),
        (("gaussian_2", "normalization"), 4),
    ]
)
def test_indices(
        model,
        path,
        index
):
    assert model.index(
        path
    ) == index


@pytest.fixture(
    name="tuple_model"
)
def make_tuple_model():
    return af.Model(MockWithTuple)


class TestTuples:
    def test_with_specific(self, tuple_model):
        with_paths = tuple_model.with_paths([
            ("tup", "0")
        ])

        assert hasattr(with_paths, "tup_0")
        assert not hasattr(with_paths, "tup_1")

    def test_without_specific(self, tuple_model):
        with_paths = tuple_model.without_paths([
            ("tup", "0")
        ])

        assert not hasattr(with_paths, "tup_0")
        assert hasattr(with_paths, "tup_1")

    def test_with(self, tuple_model):
        with_paths = tuple_model.with_paths([
            ("tup",)
        ])

        assert hasattr(with_paths, "tup")

    def test_without(self, tuple_model):
        with_paths = tuple_model.without_paths([
            ("tup",)
        ])

        assert not hasattr(with_paths, "tup")

    def test_with_explicit(self, tuple_model):
        with_paths = tuple_model.with_paths([
            ("tup", "tup_0")
        ])

        assert hasattr(with_paths, "tup_0")
        assert not hasattr(with_paths, "tup_1")

    def test_without_explicit(self, tuple_model):
        with_paths = tuple_model.without_paths([
            ("tup", "tup_0")
        ])

        assert not hasattr(with_paths, "tup_0")
        assert hasattr(with_paths, "tup_1")



import pytest

import autofit as af
from autofit.mapper.identifier import Identifier


@pytest.fixture(
    name="gaussian_prior"
)
def make_gaussian_prior():
    return Identifier(
        af.GaussianPrior(
            mean=1.0,
            sigma=2.0
        )
    )


def test_gaussian_prior_fields(
        gaussian_prior
):
    assert gaussian_prior.hash_list == [
        'GaussianPrior',
        'lower_limit',
        '-inf',
        'upper_limit',
        'inf',
        'mean',
        '1.0',
        'sigma',
        '2.0'
    ]


def test_gaussian_prior(
        gaussian_prior
):
    assert str(
        gaussian_prior
    ) == "218e05b43472cb7661b4712da640a81c"


@pytest.fixture(
    name="uniform_prior"
)
def make_uniform_prior():
    return Identifier(
        af.UniformPrior(
            lower_limit=1.0,
            upper_limit=2.0
        )
    )


def test_uniform_prior_fields(
        uniform_prior
):
    assert uniform_prior.hash_list == [
        'UniformPrior',
        'lower_limit',
        '1.0',
        'upper_limit',
        '2.0'
    ]


def test_uniform_prior(
        uniform_prior
):
    assert str(
        uniform_prior
    ) == "a0de90b9099d70b945dc56094eb5c8de"


@pytest.fixture(
    name="logarithmic_prior"
)
def make_logarithmic_prior():
    return Identifier(
        af.LogUniformPrior(
            lower_limit=1.0,
            upper_limit=2.0
        )
    )


def test_logarithmic_prior_fields(
        logarithmic_prior
):
    assert logarithmic_prior.hash_list == [
        'LogUniformPrior',
        'lower_limit',
        '1.0',
        'upper_limit',
        '2.0'
    ]


def test_logarithmic_prior(
        logarithmic_prior
):
    assert str(
        logarithmic_prior
    ) == "0e8220c88678dcb31a398f9a34dcbc8a"


@pytest.fixture(
    name="model"
)
def make_model():
    return Identifier(
        af.Model(
            af.Gaussian
        )
    )


def test_model_identifier(
        model
):
    assert str(model) == "8f342d3d4f2a60639700c295378b8b9f"


def test_model_identifier_fields(
        model
):
    assert model.hash_list == [
        'PriorModel',
        'cls',
        'autofit.example.model.Gaussian',
        'centre',
        'UniformPrior',
        'lower_limit',
        '0.0',
        'upper_limit',
        '1.0',
        'normalization',
        'UniformPrior',
        'lower_limit',
        '0.0',
        'upper_limit',
        '1.0',
        'sigma',
        'UniformPrior',
        'lower_limit',
        '0.0',
        'upper_limit',
        '1.0'
    ]



from os import path

import pytest

import autofit as af

from autofit.mock.mock_model import MockClassx3

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class TestDrawerConfig:
    def test__loads_from_config_file_correct(self):
        drawer = af.Drawer(
            total_draws=5,
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
        )

        assert drawer.prior_passer.sigma == 2.0
        assert drawer.prior_passer.use_errors is False
        assert drawer.prior_passer.use_widths is False
        assert drawer.config_dict_search["total_draws"] == 5
        assert isinstance(drawer.initializer, af.InitializerBall)
        assert drawer.initializer.lower_limit == 0.2
        assert drawer.initializer.upper_limit == 0.8
        assert drawer.number_of_cores == 1

        drawer = af.Drawer()

        assert drawer.prior_passer.sigma == 3.0
        assert drawer.prior_passer.use_errors is True
        assert drawer.prior_passer.use_widths is True
        assert drawer.config_dict_search["total_draws"] == 10
        assert isinstance(drawer.initializer, af.InitializerPrior)

    def test__samples_from_model(self):
        
        drawer = af.Drawer()
        drawer.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "drawer"))
        drawer.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx3)
        model.mock_class.one = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)
        model.mock_class.three = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)

        samples = drawer.samples_from(model=model)

        assert isinstance(samples.parameter_lists, list)
        assert isinstance(samples.parameter_lists[0], list)
        assert isinstance(samples.log_likelihood_list, list)
        assert isinstance(samples.log_prior_list, list)
        assert isinstance(samples.log_posterior_list, list)

        assert samples.parameter_lists[0] == pytest.approx(
            [49.507679, 49.177471, 14.76753], 1.0e-4
        )

        assert samples.log_likelihood_list[0] == pytest.approx(-2763.925766, 1.0e-4)
        assert samples.log_posterior_list[0] == pytest.approx(-2763.817517, 1.0e-4)
        assert samples.weight_list[0] == 1.0

        assert len(samples.parameter_lists) == 3
        assert len(samples.log_likelihood_list) == 3



import os
from pathlib import Path

import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import autofit as af
from autofit import database as db

directory = Path(__file__).parent
database_path = f"{directory}/test.db"


@pytest.fixture(
    name="session",
    scope="module"
)
def make_session():
    engine = create_engine(f"sqlite:///{database_path}")
    session = sessionmaker(bind=engine)()
    db.Base.metadata.create_all(engine)
    yield session
    session.close()
    engine.dispose()


@pytest.fixture(
    autouse=True,
    scope="module"
)
def read_in(session):
    aggregator = af.Aggregator(
        directory.parent.parent.parent.parent / "rjlens"
    )
    for item in aggregator:
        obj = db.Object.from_object(
            item.model
        )
        session.add(
            obj
        )
    yield
    try:
        os.remove(database_path)
    except FileNotFoundError:
        pass


def test_commit(session):
    session.commit()


def test_instantiate(session):
    model = session.query(
        db.Object
    ).filter(
        db.Object.parent_id.is_(None)
    ).first()
    assert isinstance(
        model(),
        af.AbstractPriorModel
    )
